! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.


!This module can be used if openmp 3.1 is not available.
module alt_add_and_fetch_module
  implicit none

  private
  
  public :: atomic_add_and_fetch

  
  interface 
     subroutine add_and_fetch_asm(value,adder,sum)
       implicit none
       integer,intent(inout) :: value
       integer,intent(in) :: adder
       integer,intent(out) :: sum
     end subroutine add_and_fetch_asm
  end interface
  
contains

  subroutine atomic_add_and_fetch(value,adder,sum)
    integer,intent(inout) :: value
    integer,intent(in) :: adder
    integer,intent(out) :: sum

    ! The following calls the fortran version when openmp is not being used,
    !  because then the compiler can use a addl instead of the asm's lock xaddl
    ! The assembly language version is used with openmp because the
    ! critical section is slower than the lock xaddl.  

    !$ if(.false.) then
       call add_and_fetch_fortran_omp2(value,adder,sum)
    !$ else
          !Call the assembly language version if using openmp
    !$   call add_and_fetch_asm(value,adder,sum)
    !$ end if
    !If you don't have an assembly language atomic add and fetch function,
    ! you either need to use the add_and_fetch_fortran function, or 
    ! write an appropriate add_and_fetch_asm function for your architecture
  end subroutine atomic_add_and_fetch

  !Example possible implementation of an atomic add_and_fetch
  subroutine add_and_fetch_fortran_omp2(value, adder, sum)
    integer,intent(inout) :: value
    integer,intent(in) :: adder
    integer,intent(out) :: sum
    
    !$OMP critical(atomic_add_and_fetch_section)
    value = value + adder
    sum = value
    !$OMP end critical(atomic_add_and_fetch_section)
  end subroutine add_and_fetch_fortran_omp2
end module alt_add_and_fetch_module
