! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

program flow_calc
  use types
  use rect_grid
  use velocity_func
  use vector_type
  use cyl_grid
  use scone_grid
  implicit none
  type(rect_grid_type) :: grid
  type(cyl_grid_type) :: cgrid
  type(scone_grid_type) :: sgrid

  call load_data("velocity_data")

  call create_grid(grid,0.2_d_t,0.2_d_t,0.2_d_t,-1.0_d_t,-1.0_d_t,-1.0_d_t,&
       10,10,16)

  call create_grid(cgrid,0.1_d_t,0.2_d_t,-1.0_d_t,6,10,16)

  call create_grid(sgrid,0.1_d_t,0.2_d_t,-1.0_d_t,6,10,16,1.0_d_t,-1.0_d_t,0.0_d_t,0.0_d_t)

  !call area_sanity_check_s(sgrid)
  
  !call calculate_all_flows(grid)
  
  !call area_sanity_check_r(grid)

  !call area_sanity_check_c(cgrid)

  !call print_points_for_zones_c(cgrid,0)
  !call print_points_for_zones_c(cgrid,688)
  !call print_points_for_zones_c(cgrid,2)
  !call print_points_for_zones_c(cgrid,690)
  
  !call calculate_all_flows_c(cgrid)

  !call print_points_for_zones_s(sgrid,9)
  !call print_points_for_zones_s(sgrid,10)
  !call print_points_for_zones_s(sgrid,12)
  !call print_points_for_zones_s(sgrid,186)
  !call print_points_for_zones_s(sgrid,42)
  !call print_points_for_zones_s(sgrid,43)
  !call print_points_for_zones_s(sgrid,204)
  !call print_points_for_zones_s(sgrid,59)
  
  !call print_points_for_all_zones_s(sgrid)

  call calculate_all_flows_s(sgrid)
contains

  subroutine print_points_for_all_zones_s(grid)
    integer :: zone,zone_id
    type(scone_grid_type),intent(in) :: grid
    zone_id = first_zone(grid)
    do zone = 1,number_of_zones(grid)
       call print_points_for_zones_s(grid,zone_id)
       zone_id = next_zone(grid,zone_id)       
    end do
  end subroutine print_points_for_all_zones_s

  subroutine print_points_for_zones_c(grid,zone_id)
    type(cyl_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id
    integer :: surface,u,v,max_u,max_v
    real(kind=d_t) :: area,x,y,z      
    type(vector) :: normal
    max_u = 10
    max_v = 10
    do surface = 1,number_of_surfaces(grid,zone_id)
       do u = 0,max_u
          do v = 0,max_v
             call get_point_normal(grid,u,v,max_u,max_v,zone_id,surface,x,y,z,area,normal)
             print *,"zone",zone_id,"surface",surface,"location",x,y,z,area,&
                  "normal",normal
          end do
       end do
    end do
  end subroutine print_points_for_zones_c

  subroutine print_points_for_zones_s(grid,zone_id)
    type(scone_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id
    integer :: surface,u,v,max_u,max_v
    real(kind=d_t) :: area,x,y,z      
    type(vector) :: normal
    max_u = 10
    max_v = 10
    do surface = 1,number_of_surfaces(grid,zone_id)
       do u = 0,max_u
          do v = 0,max_v
             call get_point_normal(grid,u,v,max_u,max_v,zone_id,surface,x,y,z,area,normal)
             if(area /= 0.0) then
                print *,"zone",zone_id,"surface",surface,&
                     "location",x,y,z,area,&
                     "normal",normal
             end if
          end do
       end do
    end do
  end subroutine print_points_for_zones_s

  subroutine area_sanity_check_c(grid)
    integer :: zone,surface,zone_id,other_zone
    type(cyl_grid_type),intent(in) :: grid
    real(kind=d_t) :: area,xc,yc,zc,area_sum
    zone_id = first_zone(grid)
    do zone = 1,number_of_zones(grid)
       do surface = 1,number_of_surfaces(grid,zone_id)
          other_zone = get_adjacent_zone(grid,zone_id,surface)
          area = area_of_surface(grid,zone_id,surface)
          area_sum = calc_area_c(grid,zone_id,surface)
          print *,"zone ",zone_id,"other",other_zone," loc ",xc,yc,zc,"area",area,"area_sum",area_sum,area/area_sum
          call print_surface_type(grid,zone_id,surface)
       end do
       call grid_zone_center(grid,zone_id,xc,yc,zc)
       zone_id = next_zone(grid,zone_id)       
    end do
  end subroutine area_sanity_check_c

  subroutine area_sanity_check_s(grid)
    integer :: zone,surface,zone_id,other_zone
    type(scone_grid_type),intent(in) :: grid
    real(kind=d_t) :: area,xc,yc,zc,area_sum
    zone_id = first_zone(grid)
    do zone = 1,number_of_zones(grid)
       do surface = 1,number_of_surfaces(grid,zone_id)
          other_zone = get_adjacent_zone(grid,zone_id,surface)
          area = area_of_surface(grid,zone_id,surface)
          area_sum = calc_area_s(grid,zone_id,surface)
          print *,"zone ",zone_id,"other",other_zone," loc ",xc,yc,zc,"area",area,"area_sum",area_sum,area/area_sum
          call print_surface_type(grid,zone_id,surface)
       end do
       call grid_zone_center(grid,zone_id,xc,yc,zc)
       zone_id = next_zone(grid,zone_id)       
    end do
  end subroutine area_sanity_check_s


  real(kind=d_t) function calc_area_c(grid,zone_id,surface)
    type(cyl_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id,surface
    integer :: u,v,umax,vmax
    real(kind=d_t) :: area,x,y,z
    type(vector) :: normal
    umax = 100
    vmax = 100
    calc_area_c = 0.0
    do u=0,umax
       do v=0,vmax
          call get_point_normal(grid,u,v,umax,vmax,zone_id,surface,x,y,z,area,normal)
          calc_area_c = calc_area_c + area
       end do
    end do
  end function calc_area_c

  real(kind=d_t) function calc_area_s(grid,zone_id,surface)
    type(scone_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id,surface
    integer :: u,v,umax,vmax
    real(kind=d_t) :: area,x,y,z
    type(vector) :: normal
    umax = 100
    vmax = 100
    calc_area_s = 0.0
    do u=0,umax
       do v=0,vmax
          call get_point_normal(grid,u,v,umax,vmax,zone_id,surface,x,y,z,area,normal)
          calc_area_s = calc_area_s + area
       end do
    end do
  end function calc_area_s


  subroutine area_sanity_check_r(grid)
    integer :: zone,surface,zone_id,other_zone
    type(rect_grid_type),intent(in) :: grid
    real(kind=d_t) :: area,xc,yc,zc,area_sum
    zone_id = first_zone(grid)
    do zone = 1,number_of_zones(grid)
       do surface = 1,number_of_surfaces(grid,zone_id)
          other_zone = get_adjacent_zone(grid,zone_id,surface)
          area = area_of_surface(grid,zone_id,surface)
          area_sum = calc_area_r(grid,zone_id,surface)
          print *,"zone ",zone_id,"other",other_zone," loc ",xc,yc,zc,"area",area,"area_sum",area_sum,area/area_sum
          !call print_surface_type(grid,zone_id,surface)
       end do
       call grid_zone_center(grid,zone_id,xc,yc,zc)
       zone_id = next_zone(grid,zone_id)       
    end do
  end subroutine area_sanity_check_r

  real(kind=d_t) function calc_area_r(grid,zone_id,surface)
    type(rect_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id,surface
    integer :: u,v,umax,vmax
    real(kind=d_t) :: area,x,y,z
    type(vector) :: normal
    umax = 100
    vmax = 100
    calc_area_r = 0.0
    do u=0,umax
       do v=0,vmax
          call get_point_normal(grid,u,v,umax,vmax,zone_id,surface,x,y,z,area,normal)
          calc_area_r = calc_area_r + area
       end do
    end do

  end function calc_area_r

  subroutine calculate_all_flows_s(grid)
    integer :: zone,surface,zone_id,other_zone
    type(scone_grid_type),intent(in) :: grid
    real(kind=d_t) :: area,flow,abs_flow,xc,yc,zc,percent
    zone_id = first_zone(grid)
    do zone = 1,number_of_zones(grid)
       flow = 0.0
       abs_flow = 0.0
       do surface = 1,number_of_surfaces(grid,zone_id)
          other_zone = get_adjacent_zone(grid,zone_id,surface)
          area = area_of_surface(grid,zone_id,surface)
          print *,"Zone ",zone_id," to ",other_zone," area ",area
          call calculate_flow_s(grid,zone_id,other_zone,surface, flow,abs_flow)
       end do
       call grid_zone_center(grid,zone_id,xc,yc,zc)
       if(abs_flow /= 0.0) then
          percent = flow/abs_flow
       else
          percent = 0.0
       end if
       print *,"zone ",zone_id," loc ",xc,yc,zc," total_flow ",flow," abs_flow ",abs_flow," percent ",percent
       zone_id = next_zone(grid,zone_id)       
    end do
  end subroutine calculate_all_flows_s

  subroutine calculate_flow_s(grid,zone_id,other_zone,surface,total_flow,abs_flow)
    type(scone_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id,other_zone,surface
    real(kind=d_t),intent(inout) :: total_flow,abs_flow
    integer :: u,v,umax,vmax
    real(kind=d_t) :: area,x,y,z,xc,yc,zc,xo,yo,zo
    type(vector) :: normal,velocity
    real(kind=d_t) :: density
    real(kind=d_t) :: flow
    umax = 100
    vmax = 100
    flow = 0.0
    do u=0,umax
       do v=0,vmax
          call get_point_normal(grid,u,v,umax,vmax,zone_id,surface,x,y,z,area,normal)
          call rect_linear_density(x,y,z,density)
          call rect_linear_interpolation(x,y,z,velocity)
          flow = flow + area * (normal .dot. velocity) 
          
       end do
    end do
    call grid_zone_center(grid,zone_id,xc,yc,zc)
    call grid_zone_center(grid,other_zone,xo,yo,zo)
    print *,"Zone ",zone_id,xc,yc,zc," to ",other_zone,xo,yo,zo," flow ",flow
    total_flow = total_flow + flow
    abs_flow = abs_flow + abs(flow)
  end subroutine calculate_flow_s


  subroutine calculate_all_flows_c(grid)
    integer :: zone,surface,zone_id,other_zone
    type(cyl_grid_type),intent(in) :: grid
    real(kind=d_t) :: area,flow,abs_flow,xc,yc,zc,percent
    zone_id = first_zone(grid)
    do zone = 1,number_of_zones(grid)
       flow = 0.0
       abs_flow = 0.0
       do surface = 1,number_of_surfaces(grid,zone_id)
          other_zone = get_adjacent_zone(grid,zone_id,surface)
          area = area_of_surface(grid,zone_id,surface)
          print *,"Zone ",zone_id," to ",other_zone," area ",area
          call calculate_flow_c(grid,zone_id,other_zone,surface, flow,abs_flow)
       end do
       call grid_zone_center(grid,zone_id,xc,yc,zc)
       if(abs_flow /= 0.0) then
          percent = flow/abs_flow
       else
          percent = 0.0
       end if
       print *,"zone ",zone_id," loc ",xc,yc,zc," total_flow ",flow," abs_flow ",abs_flow," percent ",percent
       zone_id = next_zone(grid,zone_id)       
    end do
  end subroutine calculate_all_flows_c
    
  subroutine calculate_flow_c(grid,zone_id,other_zone,surface,total_flow,abs_flow)
    type(cyl_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id,other_zone,surface
    real(kind=d_t),intent(inout) :: total_flow,abs_flow
    integer :: u,v,umax,vmax
    real(kind=d_t) :: area,x,y,z,xc,yc,zc,xo,yo,zo
    type(vector) :: normal,velocity
    real(kind=d_t) :: density
    real(kind=d_t) :: flow
    umax = 100
    vmax = 100
    flow = 0.0
    do u=0,umax
       do v=0,vmax
          call get_point_normal(grid,u,v,umax,vmax,zone_id,surface,x,y,z,area,normal)
          call rect_linear_density(x,y,z,density)
          call rect_linear_interpolation(x,y,z,velocity)
          flow = flow + area * (normal .dot. velocity) 
          
       end do
    end do
    call grid_zone_center(grid,zone_id,xc,yc,zc)
    call grid_zone_center(grid,other_zone,xo,yo,zo)
    print *,"Zone ",zone_id,xc,yc,zc," to ",other_zone,xo,yo,zo," flow ",flow
    total_flow = total_flow + flow
    abs_flow = abs_flow + abs(flow)
  end subroutine calculate_flow_c



  subroutine calculate_all_flows(grid)
    integer :: zone,surface,zone_id,other_zone
    type(rect_grid_type),intent(in) :: grid
    real(kind=d_t) :: area,flow,abs_flow,xc,yc,zc,percent
    zone_id = first_zone(grid)
    do zone = 1,number_of_zones(grid)
       flow = 0.0
       abs_flow = 0.0
       do surface = 1,number_of_surfaces(grid,zone_id)
          other_zone = get_adjacent_zone(grid,zone_id,surface)
          area = area_of_surface(grid,zone_id,surface)
          print *,"Zone ",zone_id," to ",other_zone," area ",area
          call calculate_flow(grid,zone_id,other_zone,surface, flow,abs_flow)
       end do
       call grid_zone_center(grid,zone_id,xc,yc,zc)
       if(abs_flow /= 0.0) then
          percent = flow/abs_flow
       else
          percent = 0.0
       end if
       print *,"zone ",zone_id," loc ",xc,yc,zc," total_flow ",flow," abs_flow ",abs_flow," percent ",percent
       zone_id = next_zone(grid,zone_id)       
    end do
  end subroutine calculate_all_flows
    
  subroutine calculate_flow(grid,zone_id,other_zone,surface,total_flow,abs_flow)
    type(rect_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id,other_zone,surface
    real(kind=d_t),intent(inout) :: total_flow,abs_flow
    integer :: u,v,umax,vmax
    real(kind=d_t) :: area,x,y,z,xc,yc,zc,xo,yo,zo
    type(vector) :: normal,velocity
    real(kind=d_t) :: density
    real(kind=d_t) :: flow
    umax = 100
    vmax = 100
    flow = 0.0
    do u=0,umax
       do v=0,vmax
          call get_point_normal(grid,u,v,umax,vmax,zone_id,surface,x,y,z,area,normal)
          call rect_linear_density(x,y,z,density)
          call rect_linear_interpolation(x,y,z,velocity)
          flow = flow + area * (normal .dot. velocity) 
          
       end do
    end do
    call grid_zone_center(grid,zone_id,xc,yc,zc)
    call grid_zone_center(grid,other_zone,xo,yo,zo)
    print *,"Zone ",zone_id,xc,yc,zc," to ",other_zone,xo,yo,zo," flow ",flow
    total_flow = total_flow + flow
    abs_flow = abs_flow + abs(flow)
  end subroutine calculate_flow

end program flow_calc
