! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.


PROGRAM marbles10
  ! $Id: pebbles10.f90 1378 2011-02-24 15:10:13Z cogljj $
  USE vector_type
  USE str_utils
  USE prob_table
  use number_utils
  use types
  use nearby_pebble_cache_module
  use pp_slip_module
  use geometry
  use constants
  use force_tally
  use wear_tally
  use earthquake
  use marbles9_common_module
  use network_domain_module
  use network_utils
  use version
  use new_fortran_utils
  !$ use omp_lib
  IMPLICIT NONE

  INTEGER :: i, allocate_status

  type(network_domain_type) :: network_domain

  character(256) :: line
  character(52) :: word, word1, word2, word3, word4, word5
  integer :: at
  real(kind=d_t),allocatable,dimension(:) :: temp_real_array

  integer :: number_of_proc, network_rank

  integer :: SETTINGS_UNIT
  character(32) :: ppf_base_filename = "foo_out"
  logical :: has_ppf_file = .false.
  !Should the ppf file be periodically rewound?
  logical :: rewind_ppf = .false.
  integer :: ppf_rewind_frequency = -1

  !If this is true, than boundaries are made big enough that 
  ! information only needs to be exchanged every other timestep.
  logical, parameter :: use_double_timestep_boundary = .false.

  !The interval that subprocesses use to send back information needed for 
  ! tallying.  
  integer :: tally_frequency_gcd

  real(kind=d_t) :: node_boundary_size

  integer :: rebalance_node_frequency = 100
  integer :: startup_rebalance_node_freq = 5
  integer :: startup_rebalance_length = 20

  character(1024) :: io_temp_buffer

  !Defaults
  call set_geometry_defaults
  call set_earthquake_defaults
  call set_marbles9_defaults

  call init_network()

  number_of_proc = get_network_size() 
  network_rank = get_network_rank()
  !Temporary use to force every write_time to disk.
  flush_everything = .true.

  SETTINGS_UNIT = get_settings_unit(.false.)
  input_loop: do
     read (SETTINGS_UNIT,'(A256)') line
     if(network_rank == 0) then
        print *,"#input ",trim(line)
     end if
     at = 1
     if(len_trim(line) == 0) then
        cycle input_loop
     end if
     call get_param(line,word, at)

     if(str_equal('done ',word)) then
        exit
     else if(str_equal('rem ',word)) then
        continue
     else if(str_equal('start_geometry ',word)) then
        call load_geometry_section(settings_unit,network_rank == 0)
     else if(str_equal('vessel_radius ',word)) then
        call get_param(line,word1,at)
        !inside_vessel_radius = str_to_real(word1)
        call get_param(line,word2,at)
        !outside_vessel_radius = str_to_real(word2)        
        !rectangular_vat = .false.
        call set_vessel_radius(inside=str_to_double(word1),&
             outside=str_to_double(word2))
     else if(str_equal('rectangular_vat ',word)) then
        !rectangular_vat = .true.
        call get_param(line,word1,at)
        !x_wall_distance = str_to_real(word)
        call get_param(line,word2,at)
        !y_wall_distance = str_to_real(word)
        call set_rectangular_vat(x_distance=str_to_double(word1),&
             y_distance=str_to_double(word2))
     else if(str_equal('runs ',word)) then
        call get_param(line,word,at)
        runs = str_to_int(word)
     else if(str_equal('initial_time ',word)) then
        call get_param(line,word,at)
        time = str_to_double(word)
     else if(str_equal('number_of_pebbles ',word) .or. str_equal('number_of_spheres ',word)) then
        call get_param(line,word,at)
        number_of_pebbles = str_to_int(word)
        !$ chunk_size = max(number_of_pebbles/(8*omp_get_num_procs()),1)
     else if(str_equal('kinetic_friction ',word)) then 
        call get_param(line,word,at)
        kinetic_friction = str_to_double(word)
     else if(str_equal('static_friction ',word)) then
        print *,"ERROR: static_friction is deprecated.  use static_friction_new"
        print *,"or use pebbles6"
        stop
     else if(str_equal('alpha ',word)) then
        call get_param(line,word,at)
        alpha = str_to_double(word)
     else if(str_equal('dash_pot ',word)) then
        call get_param(line,word,at)
        normal_dash_pot_constant = str_to_double(word)
        transverse_dash_pot_constant = normal_dash_pot_constant
     else if(str_equal('dash_pot2 ',word)) then
        call get_param(line,word,at)
        normal_dash_pot_constant = str_to_double(word)
        call get_param(line,word,at)
        transverse_dash_pot_constant = str_to_double(word)
     else if(str_equal('initial_packing ',word)) then        
        call get_param(line,word,at)
        initial_packing = str_to_double(word)
     else if(str_equal('random_packing_method ',word)) then
        call get_param(line,word,at)
        initial_rand_packing_mult = str_to_int(word)
     else if(str_equal('dump_frequency ',word)) then
        call get_param(line,word,at)
        dump_frequency = str_to_int(word)
     else if(str_equal('display_frequency ',word)) then
        call get_param(line,word,at)
        display_frequency = str_to_int(word)
     else if(str_equal('position_display_frequency ',word)) then
        call get_param(line,word,at)
        position_frequency = str_to_int(word)
     else if(str_equal('energy_display_frequency ',word)) then
        call get_param(line,word,at)
        energy_frequency = str_to_int(word)
     else if(str_equal('torque_enable ',word)) then 
        !torque_enable = .true.
     else if(str_equal('torque_disable ',word)) then
        print *,"ERROR: torque cannot be disabled, use pebbles6"
        !torque_enable = .false.
     else if(str_equal('seed ',word)) then
        call get_param(line,word,at)
        seed = str_to_int(word)
     else if(str_equal('earthquake_enable ',word)) then
        call get_param(line,word,at)
        !max_number_of_earthquake_waves = str_to_int(word)
        call enable_earthquake_waves(str_to_int(word))
     else if(str_equal('earthquake_sine_wave ',word)) then
        call parse_earthquake_sine_wave_line(line,at)
     else if(str_equal('earthquake_sine_wave_offset ',word)) then
        call parse_eq_sine_wave_offset_line(line,at)
     else if(str_equal('dump_positions ',word)) then
        call get_param(line,word,at)
        dump_positions_filename = word
        do_dump_positions = .true.
     else if(str_equal('dump_positions_mult ',word)) then
        call get_param(line,word,at)
        dump_positions_filename = word
        call get_param(line,word,at)
        dump_positions_mult = str_to_double(word)
        do_dump_positions = .true.
     else if(str_equal('load_positions ',word)) then
        call get_param(line,word,at)
        load_positions_filename = word
        do_load_positions = .true.
     else if(str_equal('load_positions_divide ',word)) then
        call get_param(line,word,at)
        load_positions_filename = word
        call get_param(line,word,at)
        load_positions_divide = str_to_double(word)
        do_load_positions = .true.
     else if(str_equal('load_pebble_info ',word) .or. str_equal('load_sphere_info ',word)) then
        call get_param(line,word,at)
        load_pebble_info_filename = word
        do_load_pebble_info = .true.
     else if(str_equal('load_pebble_slips ',word) .or. str_equal('load_sphere_slips ',word)) then
        do_load_pebble_slips = .true.
     else if(str_equal('load_pebble_surface_slips ',word) .or. str_equal('load_sphere_surface_slips ',word)) then
        do_load_pebble_surface_slips = .true.
     else if(str_equal('load_pebble_save ',word) .or. str_equal('load_sphere_save ',word)) then
        call get_param(line,word,at)
        load_pebble_info_filename = word
        do_load_pebble_info = .true.
        do_load_pebble_slips = .true.
        do_load_pebble_surface_slips = .true.
     else if(str_equal('floor_location ',word)) then
        call get_param(line,word,at)
        call set_floor_location(str_to_double(word))
     else if(str_equal('pebble_pebble_hooke ',word) .or. str_equal('sphere_sphere_hooke ',word)) then
        call get_param(line,word,at)
        pebble_pebble_hooke = str_to_double(word)
     else if(str_equal('pebble_vessel_hooke ',word) .or. str_equal('sphere_vessel_hooke ',word)) then
        call get_param(line,word,at)
        pebble_vessel_hooke = str_to_double(word)
     else if(str_equal('pebble_radius ',word) .or. str_equal('sphere_radius ',word)) then
        call get_param(line,word,at)
        pebble_center_radius = str_to_double(word)
        call get_param(line,word,at)
        pebble_radius = str_to_double(word)
     else if(str_equal('pebble_density ',word) .or. str_equal('sphere_density ',word)) then
        call get_param(line,word,at)
        density_center = str_to_double(word)
        call get_param(line,word,at)
        density_outside = str_to_double(word)
     else if(str_equal('static_friction_new ',word)) then
        use_static_friction = .true.
        call get_param(line,word,at)
        pp_static_friction = str_to_double(word)
        surface_static_friction = pp_static_friction
        call get_param(line,word,at)
        static_friction_hooke = str_to_double(word)
        surface_static_friction_hooke = static_friction_hooke
        call get_param(line,word,at)
        static_velocity_sqr_max = str_to_double(word)
     else if(str_equal('static_friction_new2 ',word)) then
        use_static_friction = .true.
        call get_param(line,word,at)
        pp_static_friction = str_to_double(word)
        surface_static_friction = pp_static_friction
        call get_param(line,word,at)
        static_friction_hooke = str_to_double(word)
        call get_param(line,word,at)
        surface_static_friction_hooke = str_to_double(word)
        call get_param(line,word,at)
        static_velocity_sqr_max = str_to_double(word)
     else if(str_equal('static_friction_new3 ',word)) then
        use_static_friction = .true.
        call get_param(line,word,at)
        pp_static_friction = str_to_double(word)
        call get_param(line,word,at)
        static_friction_hooke = str_to_double(word)
        call get_param(line,word,at)
        surface_static_friction = str_to_double(word)
        call get_param(line,word,at)
        surface_static_friction_hooke = str_to_double(word)
        call get_param(line,word,at)
        static_velocity_sqr_max = str_to_double(word)
     else if(str_equal('zero_long_slips ',word)) then
        zero_long_slips = .true.
        call get_param(line,word,at)  
        long_slip_sqr = str_to_double(word)**2
     else if(str_equal('decrease_long_slips ',word)) then
        decrease_long_slips = .true.
        call get_param(line,word,at)  
        decrease_slip_mult = str_to_double(word)
        call get_param(line,word,at)  
        decrease_slip_slope = str_to_double(word)
     else if(str_equal('zero_long_slips_mult ',word)) then
        zero_long_slips = .true.
        call get_param(line,word,at)  
        long_slip_mult = str_to_double(word)
     else if(str_equal('shorten_long_slips ',word)) then
        shorten_long_slips = .true.
        call get_param(line,word,at)  
        long_slip_sqr = str_to_double(word)**2
     else if(str_equal('zero_static_slips ',word)) then
        zero_static_slips = .true.
     else if(str_equal('calculate_pebble_wear ',word) .or. str_equal('calculate_sphere_wear ',word)) then
        calculate_pebble_wear_vol = .true.
        call get_param(line,word,at)
        pebble_wear_vol_display_freq = str_to_int(word)
     else if(str_equal('normal_force_wear_exponent ',word)) then
        call get_param(line,word,at)
        call set_wear_force_exponent(str_to_double(word))
     else if(str_equal('wear_velocity_bins ',word)) then
        call get_param(line,word,at)
        allocate(temp_real_array(str_to_int(word)))
        do i=1,size(temp_real_array)
           call get_param(line,word,at)
           temp_real_array(i) = str_to_double(word)
        end do
        call set_wear_velocity_bins(temp_real_array)
        deallocate(temp_real_array)
     else if(str_equal('linear_static_friction_cutoff ',word)) then
        linear_static_velocity_clamp = .true.
     else if(str_equal('no_linear_static_friction_cutoff ',word)) then
        linear_static_velocity_clamp = .false.
     else if(str_equal('pebble_of_interest ',word) .or. str_equal('sphere_of_interest ',word)) then
        call get_param(line,word,at)
        pebble_of_interest = str_to_int(word)
     else if(str_equal('use_simple_integrator ',word)) then
        integration_method = euler_method
     else if(str_equal('integration_method ',word)) then
        call get_param(line,word,at)
        if(str_equal('euler ',word)) then
           integration_method = euler_method
        else if(str_equal('runge_kutta ',word)) then
           integration_method = runge_kutta_method
        else
           print *,"ERROR unknown integration method ",word
           stop
        end if
     else if(str_equal('cone ',word)) then
        call get_param(line,word1,at)        
        !cone_location = str_to_double(word1)
        call get_param(line,word2,at)
        call set_cone(cone_location_d=str_to_double(word1), &
             cone_slope_d=str_to_double(word2))
        !cone_slope = str_to_double(word2)
     else if(str_equal('recirculate ',word)) then
        recirculate = .true.
        call get_param(line,word,at)
        call set_recirculation_height(str_to_double(word))
        call get_param(line,word,at)
        door_closed_time = str_to_double(word)
        call get_param(line,word,at)
        first_door_open_time = str_to_double(word)
        call get_param(line,word4,at)
        !hole_size = str_to_double(word)
        call get_param(line,word5,at)
        !hole_depth = str_to_double(word)
        call set_hole(hole_size_d=str_to_double(word4),&
             hole_depth_d=str_to_double(word5))
     else if(str_equal('exit_chute ',word)) then
        call get_param(line,word1,at)
        !hole_size = str_to_double(word)
        call get_param(line,word2,at)
        !hole_depth = str_to_double(word)
        call set_hole(hole_size_d=str_to_double(word1),&
             hole_depth_d=str_to_double(word2))
     else if(str_equal('recirculate_params ',word)) then
        recirculate = .true.
        call get_param(line,word,at)
        call set_recirculation_height(str_to_double(word))
        call get_param(line,word,at)
        door_closed_time = str_to_double(word)
        call get_param(line,word,at)
        first_door_open_time = str_to_double(word)
     else if(str_equal('velocity_tally ',word)) then
        call get_param(line,word,at)        
        first_velocity_tally_time = str_to_double(word)
        call get_param(line,word,at)
        velocity_tally_frequency = str_to_int(word)
     else if(str_equal('probability_table ',word)) then
        run_probability_tally = .true.
        call get_param(line,word,at)
        probability_tally_start = str_to_double(word)
        call get_param(line,word,at)
        probability_tally_frequency = str_to_int(word)
        call get_param(line,word,at)
        probability_tally_delta = str_to_double(word)
     else if(str_equal('number_of_cones ',word)) then
        call get_param(line,word,at)
        !number_of_cones = str_to_int(word)
        call set_number_of_cones(str_to_int(word))
     else if(str_equal('tabular_earthquake ',word)) then
        call parse_tabular_earthquake(line,at)
     else if(str_equal('cone_packing ',word)) then
        !use_cone_packing = .true.
        call get_param(line,word1,at)
        !packing_cone_radius = str_to_double(word)     
        call get_param(line,word2,at)
        !packing_cone_bottom = str_to_double(word)     
        call setup_cone_packing(cone_radius = str_to_double(word1),&
             cone_bottom = str_to_double(word2),chute_length = 0.0_d_t)
     else if(str_equal('cone_packing_chute ',word)) then
        !use_cone_packing = .true.
        call get_param(line,word1,at)
        !packing_cone_radius = str_to_double(word)     
        call get_param(line,word2,at)
        !packing_cone_bottom = str_to_double(word)     
        call get_param(line,word3,at)
        !packing_chute_length = str_to_double(word)
        call setup_cone_packing(cone_radius = str_to_double(word1),&
             cone_bottom = str_to_double(word2), &
             chute_length = str_to_double(word3))
     else if(str_equal('offset_inlet_holes ',word)) then
        !offset_inlet_and_cone = .true.
        call set_offset_inlet_and_cone()
     else if(str_equal('sort_pebbles ',word) .or. str_equal('sort_spheres ',word)) then
        sort_pebbles = .true.
     else if(str_equal('calculate_force_tally ',word)) then
        calculate_force_tally = .true.
     else if(str_equal('ppf_filename ',word)) then
        call get_param(line,word,at)
        ppf_base_filename = word(1:32)
        has_ppf_file = .true.
     else if(str_equal('ppf_rewind_frequency ',word)) then
        call get_param(line,word,at)
        rewind_ppf = .true.
        ppf_rewind_frequency = str_to_int(word)
     !else if(str_equal('double_timestep ',word)) then
     !   use_double_timestep_boundary = .true.
     !else if(str_equal('dump_core ',word)) then
     !   call dump_core()
     else
        print *,"ERROR unknown keyword ",word
        stop
     end if
  end do input_loop


  call open_ppf(extra_info_unit,network_rank,ppf_base_filename)
  if(.not. has_ppf_file) then
     output_suppress = .true.
  end if


  write(io_temp_buffer,*) "#Initializing_node ",network_rank," of ",number_of_proc
  call flushed_time(extra_info_unit,io_temp_buffer)


  ! Do calcuations of secondary parameters and allocations
  if(integration_method == euler_method &
       .or.  integration_method == runge_kutta_method)then 
     multistep_order = 2
  end if
  static_velocity_max = sqrt(static_velocity_sqr_max)
  long_slip = sqrt(long_slip_sqr)

  area = get_vessel_area()
  fill_edge = get_vessel_fill_edge()

  pebble_volume = (4.0/3.0)*pi*pebble_radius**3  
  height_needed = REAL(number_of_pebbles)*pebble_volume/area  
  mass = (4.0/3.0)*pi*(density_center*pebble_center_radius**3+&
       density_outside*(pebble_radius**3 - pebble_center_radius**3))
  gravitational_force = gravitational_acceleration*mass
  moment_of_inertia = 8.0/15.0 * pi * &
       (density_center * pebble_center_radius**5 + &
       density_outside * (pebble_radius ** 5 - pebble_center_radius ** 5))
  zone_increment = pebble_radius*2.0  
  call get_max_xy_extents(x_high,y_high)
  call initialize_geometry(z_low,z_high)
  z_high = z_high + height_needed/initial_packing
  x_low = -x_high
  y_low = -y_high
  if(calculate_force_tally) then
     call initialize_force_tally(x_high,x_low,y_high,y_low,z_high,z_low)
  end if
  if(use_double_timestep_boundary) then
     node_boundary_size = 5.0
  else if(integration_method == runge_kutta_method) then
     !XXX This needs to be larger than when euler_method used,
     ! but exactly how much larger is not clearly known.
     node_boundary_size = 5.2
  else
     node_boundary_size = 2.2
  end if
  call initialize_network_domain(number_of_proc,x_high,x_low,y_high,y_low,&
       z_high,z_low,network_domain,node_boundary_size*pebble_radius)

  !call debug_domain(network_domain)
  x_size = CEILING((x_high-x_low)/zone_increment)
  y_size = CEILING((y_high-y_low)/zone_increment)
  z_size = CEILING((z_high-z_low)/zone_increment)
  tally_z_inc = pebble_radius
  tally_radial_inc = pebble_radius
  rt_size = CEILING((y_high)/tally_radial_inc)
  zt_size = CEILING((z_high-z_low)/tally_z_inc)

  allocate(states(multistep_order),deltas(multistep_order))
  do i=1,multistep_order
     allocate(states(i)%pp(number_of_pebbles))
     allocate(states(i)%pv(number_of_pebbles))
     allocate(states(i)%pw(number_of_pebbles))
     ! If we don't allocate various fortran compilers act stupid.
     call allocate_pp_slip(number_of_pebbles,states(i)%pp_slip_data)
     allocate(states(i)%pss(number_of_pebbles))
     !allocate(states(i)%psn(number_of_pebbles))

     allocate(deltas(i)%pdp(number_of_pebbles))
     allocate(deltas(i)%pdv(number_of_pebbles))
     allocate(deltas(i)%pdw(number_of_pebbles))
     allocate(deltas(i)%pdss(number_of_pebbles))
     allocate(deltas(i)%keep_pss(number_of_pebbles))
     call allocate_pp_slip(number_of_pebbles,deltas(i)%pp_slip_delta)
  end do
  call initialize_pebble_cache(nearby_pebble_cache,x_size,x_low,y_size,y_low,&
       z_size,z_low,zone_increment)
  !allocate(zone(x_size,y_size,z_size,zone_count),stat=allocate_status)
  !if (allocate_status /= 0) stop "ERROR unable to allocate ram"
  !Allocate the radial tally zones
  allocate(zone_n(tally_sections,rt_size,zt_size))
  allocate(zone_vel_sum(tally_sections,rt_size,zt_size))
  allocate(zone_dot_sum(tally_sections,rt_size,zt_size))
  !Allocate the rectangular tally zones
  allocate(zone_rect_n(x_size,y_size,z_size),stat=allocate_status)
  if (allocate_status /= 0) stop "ERROR unable to allocate ram"
  allocate(zone_rect_vel_sum(x_size,y_size,z_size),stat=allocate_status)
  if (allocate_status /= 0) stop "ERROR unable to allocate ram"
  allocate(zone_rect_dot_sum(x_size,y_size,z_size),stat=allocate_status)
  if (allocate_status /= 0) stop "ERROR unable to allocate ram"

  call initialize_wear_tallies(z_low,zone_increment,z_size)
  if(run_probability_tally) then
     call initialize_count_data(probability_data,probability_tally_delta,& 
          x_low,x_high,y_low,y_high,z_low,z_high)
  end if

  time_till_door_open = first_door_open_time

  call initialize_earthquake

  !zero the tallies
  zone_n = 0
  zone_vel_sum = vector(0.0,0.0,0.0)
  zone_dot_sum = 0.0
  zone_rect_n = 0
  zone_rect_vel_sum = vector(0.0,0.0,0.0)
  zone_rect_dot_sum = 0.0

  if(use_static_friction) then 
     call clear_pp_slip(states(1)%pp_slip_data)
  end if

  tally_frequency_gcd = gcd(energy_frequency,display_frequency)
  tally_frequency_gcd = gcd(tally_frequency_gcd,position_frequency)
  tally_frequency_gcd = gcd(tally_frequency_gcd,dump_frequency)
  tally_frequency_gcd = gcd(tally_frequency_gcd,velocity_tally_frequency)
  tally_frequency_gcd = gcd(tally_frequency_gcd,probability_tally_frequency)
  !Now this is done with is_rebalance_time function
  !tally_frequency_gcd = gcd(tally_frequency_gcd,rebalance_node_frequency)



  !8 bytes for a double * 3 doubles per vector *
  !( 3 vectors for position, angular velocity, and velocity + 12 vectors
  !for slip data)) * number of pebbles.
  ! times 2 for extra transfer margin
  call create_buffer(number_of_pebbles*2*(8*3*(3+12))) 

  if(network_rank == 0) then
     call rank_zero_program()
  else
     call rank_other_program()
  end if

  ! deallocate memory
  do i=1,multistep_order
     deallocate(states(i)%pp,states(i)%pv,states(i)%pw)
     if(use_static_friction) then
        call deallocate_pp_slip(states(i)%pp_slip_data)
        deallocate(states(i)%pss) !,states(i)%psn)
     end if
     deallocate(deltas(i)%pdp,deltas(i)%pdv,deltas(i)%pdw)
  end do

  call free_earthquake

  call close_network()
CONTAINS  

  pure logical function is_full_transfer_time(run_number)
    integer,intent(in) :: run_number
    is_full_transfer_time = run_number == runs .or. &
         mod(run_number,tally_frequency_gcd) == 0 .or. &
         is_rebalance_time(run_number)

  end function is_full_transfer_time

  pure logical function is_rebalance_time(run_number)
    integer,intent(in) :: run_number    
    is_rebalance_time = (run_number < startup_rebalance_length .and. &
         mod(run_number,startup_rebalance_node_freq) == 0) .or. &
         mod(run_number,rebalance_node_frequency) == 0
    
  end function is_rebalance_time

  subroutine rank_zero_program
    !t_c = pebble multistep time index current
    !t_m1 = pebble multistep index previous
    !t_m2 = pebble multistep index for current time - alpha * 2
    integer :: t_c,t_m1,t_m2,t_m3, i
    integer, dimension(:), allocatable :: primary_pebble_list, &
         provided_pebble_list
    integer :: num_primary_pebbles, num_provided_pebbles
    real(kind=d_t) :: calc_time,start_time,end_time, time_sum
    calc_time = 1.0
    time_sum = 0.0

    call print_parameters()

    call initialize_pebble_data()

    t_c = 1

    !Calculate which nodes will get which pebbles 
    call balance_domains(number_of_proc,network_domain,states(t_c)%pp)
    call print_domain_balance(network_domain,states(t_c)%pp)

    allocate(primary_pebble_list(number_of_pebbles),provided_pebble_list(number_of_pebbles))



    call transfer_all_data(primary_pebble_list,num_primary_pebbles,&
         provided_pebble_list,num_provided_pebbles,states(t_c))


    !     Run the data    
    run_loop: DO i=0,runs     
       PRINT *,"#Run",i       
       call print_time("Start loop")

       if(rewind_ppf .and. mod(i,ppf_rewind_frequency) == 0) then
          rewind(extra_info_unit)
       end if

       write(io_temp_buffer,*) "Run",i
       call write_time(extra_info_unit,io_temp_buffer)
       start_time = get_realseconds()

       !t_m3 is index of time three steps back
       t_m3 = 1 + mod(i-2,multistep_order)
       t_m2 = 1 + mod(i-1,multistep_order)
       t_m1 = 1 + mod(i,multistep_order)
       !t_c is the current time.  states(t_c) is where the new data is written
       t_c = 1 + mod(i+1,multistep_order)

       call begin_new_force_tally_timestep()
       !Calculate only on selected ones
       if(use_double_timestep_boundary) then
          call get_simple_timestep_selected2(states(t_c),states(t_m1),&
               deltas(t_c),alpha,time, &
               provided_pebble_list,num_provided_pebbles)
         
       else
          if(integration_method == euler_method) then
             call get_simple_timestep_selected(states(t_c),states(t_m1),&
                  deltas(t_c),&
                  alpha,time, primary_pebble_list,num_primary_pebbles,&
                  provided_pebble_list,num_provided_pebbles)
          else if(integration_method == runge_kutta_method) then
             call get_rk_timestep_selected(states(t_c),states(t_m1),&
                  deltas(t_c),alpha,time,&
                  primary_pebble_list,num_primary_pebbles,&
                  provided_pebble_list,num_provided_pebbles)
          end if
       end if


       if (recirculate) then 
          call handle_recirculation_selected(states(t_c),&
               primary_pebble_list,num_primary_pebbles)
       end if

       end_time = get_realseconds()
       if(end_time > start_time) then
          !Realseconds is time since start of day, so 
          ! without the check this will fail daily.
          calc_time = end_time - start_time
       end if

       time_sum = time_sum + calc_time

       call write_time(extra_info_unit,"Transfering data")

       if((.not. use_double_timestep_boundary) .or. is_full_transfer_time(i) .or. mod(i,2) == 0) then
          call transfer_data_around(primary_pebble_list,num_primary_pebbles,&
               provided_pebble_list,num_provided_pebbles,states(t_c),&
               is_full_transfer_time(i))

          if(is_rebalance_time(i)) then
             call rebalance_nodes(time_sum,states(t_c))
             time_sum = 0.0

             call transfer_all_data(primary_pebble_list,num_primary_pebbles,&
                  provided_pebble_list,num_provided_pebbles,states(t_c))

          end if
       end if

       call do_tallies(t_c,time,i,t_m1)
       time = time + alpha
    END DO run_loop

    if(do_dump_positions) then
       call dump_positions(states(t_c)%pp,&
            dump_positions_filename,dump_positions_mult)
    end if

    if(run_probability_tally .and. time >= probability_tally_start) then
       call print_data(probability_data)
    end if

    if(calculate_force_tally) then
       call print_force_tally
    end if

    call print_date("#__end_time")
  end subroutine rank_zero_program

  subroutine transfer_data_around(primary_pebble_list,num_primary_pebbles,&
       provided_pebble_list,num_provided_pebbles,state,dump_to_zero)
    type(pebble_state),intent(inout) :: state
    integer,dimension(:),intent(inout) :: primary_pebble_list,provided_pebble_list
    integer,intent(inout) :: num_primary_pebbles,num_provided_pebbles
    !Should all the data be dumped to rank_zero node?
    logical,intent(in) :: dump_to_zero
    ! Direction(node,pebble) holds the direction that data needs to 
    !  be transfered and other information
    integer(kind=byte),dimension(:,:),allocatable :: direction
    ! SN self_node means that only can be used in node==network_rank
    ! ignore means we don't do anything special with the pebble for that node
    ! boundary_send means we send the pebble data to the node
    ! transfer_pebble means we transfer the pebble to the node
    ! primary_pebble means it was a primary pebble and it still is. SN
    ! recv_pebble means that it was transferred to this node.  SN
    ! transferred_pebble means that it was primary, but was transfered. SN
    ! boundary_recv means it is not primary, but we got it. SN
    integer(kind=byte),parameter :: IGNORE=0,BOUNDARY_SEND=1,TRANSFER_PEBBLE=2,&
         PRIMARY_PEBBLE=3,RECV_PEBBLE=4,TRANSFERRED_PEBBLE=5,BOUNDARY_RECV=6
    type(network_domain_location_type) :: location
    !Outgoing transfers and boundaries
    integer, dimension(0:number_of_proc-1) :: count_transfers,count_boundaries
    !Incoming transfers and boundaries
    integer, dimension(0:number_of_proc-1) :: incoming_transfers,incoming_boundaries
    ! data_sent is Have we sent all the data to the node, 
    ! counts_recv is do we have the counts from the node,
    ! and data_recv is have we recieved all the data
    logical, dimension(0:number_of_proc-1) :: data_sent,counts_recv,data_recv
    integer :: i,j
    integer :: incoming_rank, temp_incoming_rank
    logical :: data_available, temp_data_available
    ! Temporary lists to grab.
    integer, dimension(:),pointer :: transfer_pebble_list_recv, &
         boundary_pebble_list_recv

    ! Input data, pebble data for next state, primary pebble list
    ! Output data, new primary pebble list, new provided pebble list,
    !  updated information on boundary pebbles


    call write_time(extra_info_unit,"Figuring out transfer pebbles") 

    allocate(direction(0:number_of_proc-1,number_of_pebbles))

    !call write_time(extra_info_unit,"Initialize direction") 
    
    !direction(:,:) = IGNORE
    !$OMP PARALLEL DO
    do i=1,number_of_pebbles
       direction(:,i) = IGNORE       
    end do
    count_transfers(:) = 0
    count_boundaries(:) = 0

    !forall(i=1:num_primary_pebbles)
    !$OMP PARALLEL DO
    do i=1,num_primary_pebbles
       direction(network_rank,primary_pebble_list(i)) = PRIMARY_PEBBLE
    end do
    !end forall

    call write_time(extra_info_unit,"figure out new primary location")

    ! Figure out which pebbles need to be transmitted
    !$OMP PARALLEL DO PRIVATE(location) SCHEDULE(dynamic,chunk_size) &
    !$OMP REDUCTION(+:count_transfers,count_boundaries)
    do i=1,number_of_pebbles       
       if(direction(network_rank,i) == PRIMARY_PEBBLE) then
          location = get_location(network_domain,state%pp(i))

          ! Figure out which pebbles that we have as primary need to be 
          !  sent because they are on a boundary
          ! Note that boundary sends need to be done based on 
          !  which pebbles were primary last time, not next time
          do j = 1,number_of_boundary_domains(location)
             direction(boundary_domain(location,j),i) = BOUNDARY_SEND
             call increment(count_boundaries(boundary_domain(location,j)),1)
          end do
          ! Figure out if old primary pebble need to be transferred.
          if(primary_domain(location) /= network_rank) then
             direction(network_rank,i) = TRANSFERRED_PEBBLE
             direction(primary_domain(location),i) = TRANSFER_PEBBLE
             call increment(count_transfers(primary_domain(location)),1)
          end if
       end if
    end do


    if(dump_to_zero .and. network_rank /= 0) then
       call write_time(extra_info_unit,"dump to zero") 
       !$OMP PARALLEL DO REDUCTION(+:count_boundaries)       
       do i=1,num_primary_pebbles
          if(direction(0,primary_pebble_list(i))==IGNORE) then
             direction(0,primary_pebble_list(i)) = BOUNDARY_SEND
             call increment(count_boundaries(0),1)
          end if
       end do
    end if

    !Don't need to send boundaries to self.
    count_boundaries(network_rank) = 0
    
    write(io_temp_buffer,*) "send_data_around",network_rank,count_transfers,count_boundaries
    call write_time(extra_info_unit,io_temp_buffer)
    !Older, slower way to count transfers.
    !count_transfers = count(direction(:,:) == TRANSFER_PEBBLE,2)
    !count_boundaries = count(direction(:,:) == BOUNDARY_SEND,2)
    !write(io_temp_buffer,*) "send_data_around",network_rank,count_transfers,count_boundaries
    !call write_time(extra_info_unit,io_temp_buffer)

    call write_time(extra_info_unit,"Sending tranfer counts")

    ! Send number to update information on and number to transfer to 
    !  other nodes.
    do i=0,number_of_proc-1
       if(i /= network_rank) then
          call send_int_buffered(i, count_transfers(i))
          call send_int_buffered(i, count_boundaries(i))
       end if
    end do

    ! If no data to send, it's already sent.  
    data_sent(:) = (count_transfers(:)+count_boundaries(:)) == 0
    counts_recv(:) = .false.
    data_recv(:) = .false.

    !Already have own data    
    data_sent(network_rank) = .true.
    counts_recv(network_rank) = .true.
    data_recv(network_rank) = .true.

    call write_time(extra_info_unit,"Sending pebble data")
    do i=0,number_of_proc-1
       if(.not. data_sent(i)) then
          !We have data to send to node i.
          write(io_temp_buffer,*) "Sending data to",i,count_transfers(i),count_boundaries(i)
          call write_time(extra_info_unit,io_temp_buffer)
          
          
          ! Send transfers
          if(count_transfers(i) > 0) then
             call send_pebble_byte_match_list(i,direction(i,:),TRANSFER_PEBBLE,count_transfers(i),state)
          end if
          if(count_boundaries(i) > 0) then
             call send_pebble_byte_match_list(i,direction(i,:),BOUNDARY_SEND,count_boundaries(i),state)
          end if
          data_sent(i) = .true.
          
          write(io_temp_buffer,*) "Sent data to",i
          call write_time(extra_info_unit,io_temp_buffer)
          
       end if
    end do

    call write_time(extra_info_unit,"Waiting for other data")


    ! While we still have data to send or recieve, 
    transfer_data_loop: do 
       if(count(.not. data_sent(:)) + count(.not. data_recv(:)) == 0) then
          ! When data_sent is all true and data_recv is all true exit
          exit transfer_data_loop
       end if

       !check for incoming messages from nodes we are waiting on data_recv for
       data_available = .false.
       check_recv: do i=0,number_of_proc-1
          if(.not. data_recv(i)) then
             call iprobe_network_recv(i,temp_incoming_rank,temp_data_available)
             if(temp_data_available) then
                data_available = .true.
                incoming_rank = temp_incoming_rank
                exit check_recv
             end if
          end if
       end do check_recv

       if(data_available) then
          !If there is an incoming transfer and layer sequence recieve it
          if(counts_recv(incoming_rank)) then
             write(io_temp_buffer,*) "Getting data from",incoming_rank,&
                  incoming_transfers(incoming_rank),incoming_boundaries(incoming_rank)
             call write_time(extra_info_unit,io_temp_buffer)

             ! This must be the data
             if(incoming_transfers(incoming_rank) > 0) then
                call recv_pebble_list(incoming_rank, state, transfer_pebble_list_recv)
             end if
             if(incoming_boundaries(incoming_rank) > 0) then
                call recv_pebble_list(incoming_rank, state, boundary_pebble_list_recv)
             end if
             data_recv(incoming_rank) = .true.

             write(io_temp_buffer,*) "recv_data_around",network_rank,&
                  incoming_rank,size(transfer_pebble_list_recv),size(boundary_pebble_list_recv)
             call write_time(extra_info_unit,io_temp_buffer)

             !Mark new pebbles as recieved
             if(incoming_transfers(incoming_rank) > 0) then
                forall(i=1:size(transfer_pebble_list_recv))
                   direction(network_rank,transfer_pebble_list_recv(i)) = &
                        RECV_PEBBLE
                end forall
                deallocate(transfer_pebble_list_recv)
             end if
             if(incoming_boundaries(incoming_rank) > 0) then
                forall(i=1:size(boundary_pebble_list_recv))
                   direction(network_rank,boundary_pebble_list_recv(i)) = &
                        BOUNDARY_RECV
                end forall
                deallocate(boundary_pebble_list_recv)
             end if
          else 
             write(io_temp_buffer,*) "Getting counts from",incoming_rank
             call write_time(extra_info_unit,io_temp_buffer)

             ! This must be the transfer counts
             incoming_transfers(incoming_rank) =recv_network_int(incoming_rank)
             incoming_boundaries(incoming_rank)=recv_network_int(incoming_rank)
             counts_recv(incoming_rank) = .true.
             !If no transfers or boundaries, then done with data too.
             data_recv(incoming_rank) = (incoming_transfers(incoming_rank) + incoming_boundaries(incoming_rank) <= 0)

             write(io_temp_buffer,*) "Got counts from",incoming_rank,&
                  incoming_transfers(incoming_rank),&
                  incoming_boundaries(incoming_rank)
             call write_time(extra_info_unit,io_temp_buffer)
          end if
       end if


    end do transfer_data_loop


    call write_time(extra_info_unit,"Finished transfering data")

    !This is not strictly necessary if the MPI implementation is correct
    ! MVAPICH2 version 1.2p1 requires this for optimal speed however.
    call network_buffer_flush()

    call write_time(extra_info_unit,"Flushed network buffers data")


    !update provided lists and primary pebble lists

    num_primary_pebbles = 0
    num_provided_pebbles = 0
    do i=1,number_of_pebbles
       if(direction(network_rank,i) == PRIMARY_PEBBLE .or. &
            direction(network_rank,i) == RECV_PEBBLE) then
          num_primary_pebbles = num_primary_pebbles + 1
          num_provided_pebbles = num_provided_pebbles + 1
          primary_pebble_list(num_primary_pebbles) = i
          provided_pebble_list(num_provided_pebbles) = i
       else if(direction(network_rank,i) == BOUNDARY_RECV) then
          num_provided_pebbles = num_provided_pebbles + 1
          provided_pebble_list(num_provided_pebbles) = i
       end if
    end do

    !if(.not. allocated(direction)) write(extra_info_unit,*) "direction not allocated"
    deallocate(direction)
  end subroutine transfer_data_around

  subroutine transfer_all_data(primary_pebble_list,num_primary_pebbles,&
       provided_pebble_list,num_provided_pebbles,state)
    type(pebble_state),intent(inout) :: state
    integer,dimension(:),intent(inout) :: primary_pebble_list,provided_pebble_list
    integer,intent(inout) :: num_primary_pebbles,num_provided_pebbles
    integer, dimension(:),pointer :: domain_transfer_array, &
         primary_pebble_list_recv, boundary_pebble_list_recv
    ! boundary(node,pebble) is true if the pebble is in the boundary layer of 
    !  for that node
    ! primary(node,pebble) is true if the pebble is owned by that node
    logical, dimension(:,:),allocatable :: boundary_pebbles, primary_pebbles
    integer, dimension(:),allocatable :: count_transfers,count_boundaries
    type(network_domain_location_type) :: location

    
    integer :: length, source, num_boundary_pebbles,j,k
    
    if(network_rank == 0) then
       
       domain_transfer_array => get_integer_data_save(network_domain)
       do i  = 1,number_of_proc-1
          call send_int_array(i, domain_transfer_array,size(domain_transfer_array))
       end do

       !Distribute initial data to all the cluster nodes
       allocate(boundary_pebbles(0:number_of_proc-1,number_of_pebbles))
       allocate(primary_pebbles(0:number_of_proc-1,number_of_pebbles))
       allocate(count_transfers(0:number_of_proc-1),count_boundaries(0:number_of_proc-1))

       boundary_pebbles(:,:) = .false.
       primary_pebbles(:,:) = .false.

       do i = 1,number_of_pebbles
          location = get_location(network_domain,state%pp(i))
          primary_pebbles(primary_domain(location),i) = .true.
          forall(j = 1:number_of_boundary_domains(location))
             boundary_pebbles(boundary_domain(location,j),i) = .true.
          end forall
          if(is_pebble_of_interest(i)) then
             write(extra_info_unit,*) "#POID",i,"d",primary_domain(location),"nb",number_of_boundary_domains(location)
          end if
       end do

       count_transfers = count(primary_pebbles,2)
       count_boundaries = count(boundary_pebbles,2)
       do i = 1,number_of_proc-1
          call send_int(i, count_transfers(i))
          call send_int(i,count_boundaries(i))
       end do

       do i = 1,number_of_proc-1
          call send_pebble_list(i, primary_pebbles(i,:),state)
       end do

       do i = 1,number_of_proc-1
          call send_pebble_list(i, boundary_pebbles(i,:),state)
       end do

       num_primary_pebbles = count(primary_pebbles(network_rank,:))
       num_boundary_pebbles = count(boundary_pebbles(network_rank,:))
       num_provided_pebbles = num_primary_pebbles + num_boundary_pebbles

       j = 1
       k = 1
       do i=1,number_of_pebbles
          if(primary_pebbles(network_rank,i)) then
             primary_pebble_list(j) = i
             provided_pebble_list(k) = i
             j = j + 1
             k = k + 1
          else if(boundary_pebbles(network_rank,i)) then
             provided_pebble_list(k) = i
             k = k+1
          end if
       end do
 
       !if(.not. allocated(boundary_pebbles)) write(extra_info_unit,*) "boundary_pebbles not allocated"
       !if(.not. allocated(primary_pebbles)) write(extra_info_unit,*) "primary_pebbles not allocated"
       !if(.not. allocated(count_transfers)) write(extra_info_unit,*) "count_transfers not allocated"
       !if(.not. allocated(count_boundaries)) write(extra_info_unit,*) "count_boundaries not allocated"
       deallocate(boundary_pebbles,primary_pebbles,count_transfers,count_boundaries)
    else ! Not network rank zero
       !Receieve network domain information
       !Data should come from node 0, otherwise it is possible to get data from somewhere else
       call probe_integer_recv(0, source, length)
       if(source /= 0) then 
          print *,"#ERROR ",network_rank," expected data from 0, but got data from ",source
       end if
       allocate(domain_transfer_array(length))
       call recv_network_int_array(source, domain_transfer_array, length)
       call load_domains_from_integer_array(network_domain,domain_transfer_array)

       num_primary_pebbles = recv_network_int(source)
       num_boundary_pebbles = recv_network_int(source)

       ! primary pebbles
       call recv_pebble_list(source, state, primary_pebble_list_recv)
       ! boundary pebbles
       call recv_pebble_list(source, state, boundary_pebble_list_recv)

       primary_pebble_list(1:num_primary_pebbles) = primary_pebble_list_recv
       call union_pebble_lists(primary_pebble_list_recv,num_primary_pebbles,&
            boundary_pebble_list_recv,num_boundary_pebbles,&
            provided_pebble_list,num_provided_pebbles)
       if(has_ppf_file) then
          write(extra_info_unit,*) "#upl",num_primary_pebbles,num_boundary_pebbles,num_provided_pebbles,number_of_pebbles
       end if

       deallocate(primary_pebble_list_recv,boundary_pebble_list_recv)
    end if
    deallocate(domain_transfer_array)

  end subroutine transfer_all_data

  subroutine rebalance_nodes(time_sum,state)
    real(kind=d_t),intent(in) :: time_sum
    type(pebble_state),intent(in) :: state
    real(kind=d_t),dimension(:),allocatable :: work

    if(network_rank == 0) then
       allocate(work(0:number_of_proc-1))
       work(0) = time_sum
       do i=1,number_of_proc-1
          work(i) = recv_network_double(i)
       end do
       call balance_domains_work(network_domain,state%pp,work)
       call print_domain_balance(network_domain,state%pp)
       !if(.not. allocated(work)) write(extra_info_unit,*) "no work allocated"
       deallocate(work)
    else ! Not network rank zero
       call send_double(0,time_sum)
       
    end if

  end subroutine rebalance_nodes

  subroutine do_tallies(t_c,time,i,t_m1)
    integer,intent(in) :: t_c,i,t_m1
    real(kind=d_t),intent(in) :: time

    call handle_velocity_tally(states(t_c),time,i)
    
    if(run_probability_tally .and. time >= probability_tally_start &
         .and. MOD(i,probability_tally_frequency) .eq. 0) then
       call handle_probability_tally(states(t_c),states(t_m1))
    end if
    
    IF(MOD(i,display_frequency) .eq. 0 .or. i == runs) THEN
       PRINT *,"#Display Run",i
       call display_pebble_info(time,states(t_c))
    else if(mod(i,energy_frequency) .eq. 0) then
       call display_pebble_energy(time,states(t_c))
    END IF
    if(mod(i,position_frequency) .eq. 0) then
       call display_pebble_positions(time,states(t_c))
    end if
    if(mod(i,dump_frequency) .eq. 0 .or. i == runs) then
       call display_overlaps(states(t_c)%pp)
       call display_zone_velocities()
    end if
    if(any_pebbles_of_interest()) then
       call dump_pebbles_of_interest_info(states(t_c),deltas(t_c),time)
    end if
    if(calculate_pebble_wear_vol .and. mod(i,pebble_wear_vol_display_freq) == 0) then
       call print_pebble_wear_data(time,alpha)
    end if
  end subroutine do_tallies

  subroutine rank_other_program
    !t_c = pebble multistep time index current
    !t_m1 = pebble multistep index previous
    !t_m2 = pebble multistep index for current time - alpha * 2
    integer :: t_c,t_m1,t_m2,t_m3, i
    integer, dimension(:), allocatable :: primary_pebble_list, &
         provided_pebble_list
    integer :: num_primary_pebbles, num_provided_pebbles
    real(kind=d_t) :: calc_time,start_time,end_time,time_sum
    calc_time = 1.0
    time_sum = 0.0


    t_c = 1
    
    allocate(primary_pebble_list(number_of_pebbles))
    allocate(provided_pebble_list(number_of_pebbles))
    
    call transfer_all_data(primary_pebble_list,num_primary_pebbles,&
         provided_pebble_list,num_provided_pebbles,states(t_c))

    !     Run the data    
    run_loop: DO i=0,runs     
       write(io_temp_buffer,*) "Run",i
       if(rewind_ppf .and. mod(i,ppf_rewind_frequency) == 0) then
          rewind(extra_info_unit)
       end if

       call write_time(extra_info_unit,io_temp_buffer)
       start_time = get_realseconds()

       !t_m3 is index of time three steps back
       t_m3 = 1 + mod(i-2,multistep_order)
       t_m2 = 1 + mod(i-1,multistep_order)
       t_m1 = 1 + mod(i,multistep_order)
       !t_c is the current time.  states(t_c) is where the new data is written
       t_c = 1 + mod(i+1,multistep_order)

       if(use_double_timestep_boundary) then
          call get_simple_timestep_selected2(states(t_c),states(t_m1),&
               deltas(t_c),alpha,time, &
               provided_pebble_list,num_provided_pebbles)
       else
          if(integration_method == euler_method) then
             call get_simple_timestep_selected(states(t_c),states(t_m1),&
                  deltas(t_c),&
                  alpha,time, primary_pebble_list,num_primary_pebbles,&
                  provided_pebble_list,num_provided_pebbles)
          else if(integration_method == runge_kutta_method) then
             call get_rk_timestep_selected(states(t_c),states(t_m1),&
                  deltas(t_c),alpha,time,&
                  primary_pebble_list,num_primary_pebbles,&
                  provided_pebble_list,num_provided_pebbles)
          end if
       end if

       if (recirculate) then 
          call handle_recirculation_selected(states(t_c),&
               primary_pebble_list,num_primary_pebbles)
       end if

       end_time = get_realseconds()
       if(end_time > start_time) then
          !Realseconds is time since start of day, so 
          ! without the check this will fail daily.
          calc_time = end_time - start_time
       end if

       time_sum = time_sum + calc_time

       call write_time(extra_info_unit,"Transfering data")
       
       if((.not. use_double_timestep_boundary) .or. is_full_transfer_time(i) .or. mod(i,2) == 0) then
          call transfer_data_around(primary_pebble_list,num_primary_pebbles,&
            provided_pebble_list,num_provided_pebbles,states(t_c),&
            is_full_transfer_time(i))

          if(is_rebalance_time(i)) then
             call rebalance_nodes(time_sum,states(t_c))
             time_sum = 0.0
             
             call transfer_all_data(primary_pebble_list,num_primary_pebbles,&
                  provided_pebble_list,num_provided_pebbles,states(t_c))
             
          end if
       end if

       if(any_pebbles_of_interest()) then
          call dump_pebbles_of_interest_info(states(t_c),deltas(t_c),time)
       end if

       time = time + alpha

    end DO run_loop

    !call display_pebble_energy(time,states(t_c))

  end subroutine rank_other_program

  subroutine print_parameters() 
    PRINT *,"#Pebbles"  
    PRINT *,"#Version $Id: pebbles10.f90 1378 2011-02-24 15:10:13Z cogljj $"
    call print_version    
    PRINT *,"# Prepared by Battelle Energy Alliance, LLC"
    PRINT *,"# Under Contract No. DE-AC07-05ID14517"
    PRINT *,"# With the U. S. Department of Energy"
    PRINT *,"#Pebble_Count: ",number_of_pebbles  
    PRINT *,"#alpha",alpha," grav",gravitational_force  
    PRINT *,"#runs",runs,runs*alpha,"s long",time,"s start"
    PRINT *,"#sizes",x_size,y_size,z_size  
    PRINT *,"#sp ss k friction",pp_static_friction,surface_static_friction,kinetic_friction  
    if(use_static_friction) then
       PRINT *,"#static velocity max",static_velocity_max
       PRINT *,"#static friction hooke",static_friction_hooke, surface_static_friction_hooke
       PRINT *,"#linear static velocity",linear_static_velocity_clamp,zero_static_slips
    else 
       PRINT *,"#Not using static friction"
    end if
    PRINT *,"#n_dashpot",normal_dash_pot_constant," t_dashpot ",&
         transverse_dash_pot_constant," air_friction ",air_friction  
    PRINT *,"#hooke", pebble_pebble_hooke, pebble_vessel_hooke  
    PRINT *,"#d_t",d_t
    PRINT *,"#display_frequency",display_frequency,display_frequency*alpha
    PRINT *,"#mass",mass," moment_of_inertia" ,moment_of_inertia
    call print_geometry_info()
    if(recirculate) then 
       PRINT *,"#door_closed_time",door_closed_time,"ttdo",time_till_door_open
    else 
       PRINT *,"#recirculation off"
    end if
    PRINT *,"#first_velocity_tally_time",first_velocity_tally_time,"vtf",&
         velocity_tally_frequency
    call print_earthquake_info
    PRINT *,"#torque_enable",torque_enable
    PRINT *,"#static_friction",use_static_friction,"spf",pp_static_friction,&
         "ssf",surface_static_friction,&
         "hooke",static_friction_hooke,"v**2",static_velocity_sqr_max,&
         "surface_hooke",surface_static_friction_hooke
    PRINT *,"#display_frequency",display_frequency,display_frequency*alpha
    PRINT *,"#run_length",runs,runs*alpha
    PRINT *,"#seed",seed
    PRINT *,"#use_simple_integrator",integration_method == euler_method
    print *,"#sizes",x_size,y_size,z_size
    print *,"#near_pebble_size",near_pebbles_size
    !$ print *,"#omp_procs",omp_get_num_procs()
    print *,"#tally_frequency_gcd",tally_frequency_gcd
    call print_date("#start_time")
  end subroutine print_parameters

  subroutine union_pebble_lists(list_a,list_a_length,list_b,list_b_length,&
       union_list,union_list_length)
    integer, intent(in) :: list_a_length,list_b_length
    integer,intent(out) :: union_list_length
    integer,dimension(:),intent(in) :: list_a,list_b
    integer,dimension(:),intent(out) :: union_list
    logical,dimension(:),allocatable :: list_a_logical,list_b_logical
    integer :: i
    allocate(list_a_logical(number_of_pebbles),&
         list_b_logical(number_of_pebbles))
    list_a_logical = .false.
    list_b_logical = .false.
    do i=1,list_a_length
       list_a_logical(list_a(i)) = .true.
    end do
    do i=1,list_b_length
       list_b_logical(list_b(i)) = .true.
    end do
    union_list_length = 0
    do i=1,number_of_pebbles
       if(list_a_logical(i) .or. list_b_logical(i)) then
          union_list_length = union_list_length + 1
          union_list(union_list_length) = i
       end if
    end do
    !if(.not. allocated(list_a_logical) .or. .not. allocated(list_b_logical)) then
    !   write(extra_info_unit,*) "list_a_logical or list_b_logical not allocated"
    !end if
    deallocate(list_a_logical,list_b_logical)
  end subroutine union_pebble_lists
  

  subroutine recv_pebble_list(source, state, pebble_list_pointer)
    integer, intent(in) :: source
    type(pebble_state), intent(inout) :: state
    integer, dimension(:), pointer, optional :: pebble_list_pointer
    integer, dimension(:), pointer :: pebble_list
    integer :: pebble_list_len, pp_int_array_len,pp_double_array_len, p_i
    integer :: temp_real_mult,list_adder, t_i
    real(kind=d_t),dimension(:),allocatable :: temp_reals
    integer, dimension(:),pointer :: pp_int_array
    real(kind=d_t),dimension(:),pointer :: pp_double_array


    !call write_time(extra_info_unit,"Recieving pebble_list_len")
    pebble_list_len = recv_network_int(source)
    !call write_time(extra_info_unit,"allocating")
    allocate(pebble_list(pebble_list_len))
    
    if(use_static_friction) then
       !Number of pebbles * (x,y,z) * (position,linear,angular,sslip)
       temp_real_mult = 3*4
    else
       !Number of pebbles * (x,y,z) * (position,linear,angular)
       temp_real_mult = 3*3
    end if
    allocate(temp_reals(number_of_pebbles*temp_real_mult))
    
    !call write_time(extra_info_unit,"Recieving pebble_list")
    call recv_network_int_array(source,pebble_list,pebble_list_len)
    
    !call write_time(extra_info_unit,"Recieving large array")
    !positions
    call recv_network_double_array(source,temp_reals,&
         pebble_list_len*temp_real_mult)
    ! The OMP sections don't seem to do that much speedup.
    !$OMP SECTIONS private(i,p_i,list_adder,t_i)
    !$OMP SECTION
    !call write_time(extra_info_unit,"Recieving positions")
    list_adder = 0
    do i=1,pebble_list_len
       p_i = pebble_list(i)
       t_i = 1+(i-1)*3+list_adder
       state%pp(p_i)%x = temp_reals(t_i+0)
       state%pp(p_i)%y = temp_reals(t_i+1)
       state%pp(p_i)%z = temp_reals(t_i+2)
    end do
    !call write_time(extra_info_unit,"Done Recieving positions")

    !call write_time(extra_info_unit,"Recieving linear velocities")
    !linear velocities
    list_adder = 3*pebble_list_len
    !call recv_network_double_array(source,temp_reals,pebble_list_len)
    do i=1,pebble_list_len
       p_i = pebble_list(i)
       t_i = 1+(i-1)*3+list_adder       
       state%pv(p_i)%x = temp_reals(t_i+0)
       state%pv(p_i)%y = temp_reals(t_i+1)
       state%pv(p_i)%z = temp_reals(t_i+2)       
    end do

    !call write_time(extra_info_unit,"Recieving angular velocities")
    ! angular velocities
    list_adder = 6*pebble_list_len
    !call recv_network_double_array(source,temp_reals,pebble_list_len)
    do i=1,pebble_list_len
       p_i = pebble_list(i)
       t_i = 1+(i-1)*3+list_adder       
       state%pw(p_i)%x = temp_reals(t_i+0)
       state%pw(p_i)%y = temp_reals(t_i+1)
       state%pw(p_i)%z = temp_reals(t_i+2)
    end do

    if(use_static_friction) then

       !call write_time(extra_info_unit,"Recieving pebble surface slips")
       !pebble surface slips
       !call recv_network_double_array(source,temp_reals,pebble_list_len)
       list_adder = 9*pebble_list_len
       do i=1,pebble_list_len
          p_i = pebble_list(i)
          t_i = 1+(i-1)*3+list_adder       
          state%pss(p_i)%x = temp_reals(t_i+0)
          state%pss(p_i)%y = temp_reals(t_i+1)
          state%pss(p_i)%z = temp_reals(t_i+2)
       end do

    end if
    !call write_time(extra_info_unit,"Done processing large array data")

    !$OMP SECTION

    if(use_static_friction) then

       !call write_time(extra_info_unit,"Recieving pebble pebble slips")
       ! pebble pebble slips
       !call probe_integer_recv(source,source_probe, probe_length)
       pp_int_array_len = recv_network_int(source)
       pp_double_array_len = recv_network_int(source)

       if(pp_int_array_len > 0) then
          allocate(pp_int_array(pp_int_array_len))
       
          call recv_network_int_array(source,pp_int_array,pp_int_array_len)
       
          allocate(pp_double_array(pp_double_array_len))
          call recv_network_double_array(source,pp_double_array,&
               pp_double_array_len)
       
          call load_from_partial_pp_slip_dump(state%pp_slip_data,&
               pp_int_array,pp_double_array)

          deallocate(pp_int_array,pp_double_array)

       end if

       !call write_time(extra_info_unit,"Done recieving pebble pebble slips")
    end if

    !$OMP END SECTIONS

    
    !call write_time(extra_info_unit,"done with recv pebble list")
    if(present(pebble_list_pointer)) then
       pebble_list_pointer => pebble_list
    else
       deallocate(pebble_list)
    end if
    !if(.not. allocated(temp_reals) .or. .not. allocated(temp_ints)) then 
    !   write(extra_info_unit,*) "temp_reals or temp_ints not allocated rpl"
    !end if
    deallocate(temp_reals)
  end subroutine recv_pebble_list

  !This version takes a byte list of which to send.
  subroutine send_pebble_byte_match_list(destination,pebble_send_list, match, pebble_list_len, state)
    integer, intent(in) :: destination, pebble_list_len
    integer(kind=byte), intent(in), dimension(:) :: pebble_send_list
    integer(kind=byte), intent(in) :: match
    type(pebble_state),intent(in):: state
    integer, allocatable,dimension(:) :: pebble_list
    !integer :: pebble_list_len
    integer :: i, j

    !pebble_list_len = count(pebble_send_list)

    allocate(pebble_list(pebble_list_len))
    
    j = 1
    do i=1,number_of_pebbles
       if(pebble_send_list(i)==match) then
          pebble_list(j) = i
          j = j + 1
          if(is_pebble_of_interest(i)) then
             write(extra_info_unit,*) "#POISEND",i,"from",network_rank,"to",destination,&
                  "p",state%pp(i),"v",state%pv(i),"w",state%pw(i)
          end if
       end if
    end do

    call send_pebble_list2(destination,pebble_list,pebble_list_len,state)

    deallocate(pebble_list)    
  end subroutine send_pebble_byte_match_list


  !This version takes a logical list of which to send.
  subroutine send_pebble_list(destination,pebble_send_list, state)
    integer, intent(in) :: destination
    logical, intent(in), dimension(:) :: pebble_send_list
    type(pebble_state),intent(in):: state
    integer, allocatable,dimension(:) :: pebble_list
    integer :: pebble_list_len
    integer :: i, j

    pebble_list_len = count(pebble_send_list)

    allocate(pebble_list(pebble_list_len))
    
    j = 1
    do i=1,number_of_pebbles
       if(pebble_send_list(i)) then
          pebble_list(j) = i
          j = j + 1
          if(is_pebble_of_interest(i)) then
             write(extra_info_unit,*) "#POISEND",i,"from",network_rank,"to",destination,&
                  "p",state%pp(i),"v",state%pv(i),"w",state%pw(i)
          end if
       end if
    end do

    call send_pebble_list2(destination,pebble_list,pebble_list_len,state)

    deallocate(pebble_list)    
  end subroutine send_pebble_list

  !This version takes a list and a length
  subroutine send_pebble_list2(destination,pebble_list,pebble_list_len,state)
    integer, intent(in) :: destination, pebble_list_len
    integer, dimension(:),intent(in) :: pebble_list
    type(pebble_state),intent(in):: state
    real(kind=d_t),dimension(:),allocatable :: temp_reals
    integer, dimension(:),pointer :: pp_int_array
    real(kind=d_t),dimension(:),pointer :: pp_double_array
    integer :: p_i,t_i, i, list_adder, temp_real_mult

    !call write_time(extra_info_unit,"Entering send_pebble_list2")

    if(use_static_friction) then
       !Number of pebbles * (x,y,z) * (position,linear,angular,sslip)
       temp_real_mult = 3*4
    else
       !Number of pebbles * (x,y,z) * (position,linear,angular)
       temp_real_mult = 3*3
    end if

    allocate(temp_reals(number_of_pebbles*temp_real_mult))
    
    call send_int_buffered(destination,pebble_list_len)
    call send_int_array_buffered(destination,pebble_list,pebble_list_len)

    list_adder = 0
    ! positions
    do i=1,pebble_list_len
       p_i = pebble_list(i)
       t_i = 1+(i-1)*3+list_adder
       temp_reals(t_i+0) = state%pp(p_i)%x
       temp_reals(t_i+1) = state%pp(p_i)%y
       temp_reals(t_i+2) = state%pp(p_i)%z
    end do
   
    !call send_double_array_buffered(destination,temp_reals,pebble_list_len)
    list_adder = list_adder + pebble_list_len*3

    ! linear velocities
    do i=1,pebble_list_len
       p_i = pebble_list(i)
       t_i = 1+(i-1)*3+list_adder
       temp_reals(t_i+0) = state%pv(p_i)%x
       temp_reals(t_i+1) = state%pv(p_i)%y
       temp_reals(t_i+2) = state%pv(p_i)%z
    end do
    list_adder = list_adder + pebble_list_len*3
    !call send_double_array_buffered(destination,temp_reals,pebble_list_len)

    ! angular velocities
    do i=1,pebble_list_len
       p_i = pebble_list(i)
       t_i = 1+(i-1)*3+list_adder
       temp_reals(t_i+0) = state%pw(p_i)%x
       temp_reals(t_i+1) = state%pw(p_i)%y
       temp_reals(t_i+2) = state%pw(p_i)%z
    end do
    list_adder = list_adder + pebble_list_len*3

    !call send_double_array_buffered(destination,temp_reals,pebble_list_len)

    if(use_static_friction) then
       !pebble surface slips
       do i=1,pebble_list_len
          p_i = pebble_list(i)
          t_i = 1+(i-1)*3+list_adder
          temp_reals(t_i+0) = state%pss(p_i)%x
          temp_reals(t_i+1) = state%pss(p_i)%y
          temp_reals(t_i+2) = state%pss(p_i)%z
       end do

       call send_double_array_buffered(destination,temp_reals,&
            pebble_list_len*temp_real_mult)       
       
       call get_partial_pp_slip_dump(pebble_list,pebble_list_len, &
            state%pp_slip_data, pp_int_array, pp_double_array)
       call send_int_buffered(destination,size(pp_int_array))
       call send_int_buffered(destination,size(pp_double_array))
       
       call send_int_array_buffered(destination, pp_int_array,size(pp_int_array))
       call send_double_array_buffered(destination,pp_double_array,size(pp_double_array))
       deallocate(pp_int_array,pp_double_array)
    else
       call send_double_array_buffered(destination,temp_reals,&
            pebble_list_len*temp_real_mult)       
       
    end if

    !if(.not. allocated(pebble_list)) write(extra_info_unit,*) "pebble_list not allocated spl"
    deallocate(temp_reals)

    !call write_time(extra_info_unit,"Exiting send_pebble_list2")

  end subroutine send_pebble_list2

  ! Opens a file depending on the marker given
  subroutine open_ppf(unit,marker,filename)
    integer,intent(in) :: marker
    integer,intent(inout) :: unit
    character(32),intent(in) :: filename
    character(60) :: filename_copy
    write(filename_copy,'(A,I4.4)') trim(filename),marker
    open(unit=unit,file=filename_copy,status="REPLACE")

  end subroutine open_ppf

  subroutine dump_core()
    integer,pointer :: ip
    nullify(ip)
    ip = 1  !Assign to a null pointer, to cause a segmentation fault.
  end subroutine dump_core

END PROGRAM marbles10
