! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.


PROGRAM marbles9
  ! $Id: pebbles9.f90 1378 2011-02-24 15:10:13Z cogljj $
  USE vector_type
  USE str_utils
  USE prob_table
  use number_utils
  use types
  use nearby_pebble_cache_module
  use pp_slip_module
  use geometry
  use constants
  use force_tally
  use earthquake
  use marbles9_common_module
  use version
  !$ use omp_lib
  IMPLICIT NONE

  INTEGER :: i, allocate_status, settings_unit

  character(256) :: line
  character(52) :: word, word1, word2, word3, word4, word5
  integer :: at
  real(kind=d_t),allocatable,dimension(:) :: temp_real_array

  !Defaults
  call set_geometry_defaults
  call set_earthquake_defaults
  call set_marbles9_defaults
  extra_info_unit = stdout_unit

  settings_unit = get_settings_unit(.true.)
  input_loop: do
     read (settings_unit,'(A256)') line
     print *,"#input ",trim(line)
     at = 1
     if(len_trim(line) == 0) then
        cycle input_loop
     end if
     call get_param(line,word, at)

     if(str_equal('done ',word)) then
        exit
     else if(str_equal('rem ',word)) then
        continue
     else if(str_equal('start_geometry ',word)) then
        call load_geometry_section(settings_unit,.true.)
     else if(str_equal('vessel_radius ',word)) then
        call get_param(line,word1,at)
        !inside_vessel_radius = str_to_real(word1)
        call get_param(line,word2,at)
        !outside_vessel_radius = str_to_real(word2)        
        !rectangular_vat = .false.
        call set_vessel_radius(inside=str_to_double(word1),&
             outside=str_to_double(word2))
     else if(str_equal('rectangular_vat ',word)) then
        !rectangular_vat = .true.
        call get_param(line,word1,at)
        !x_wall_distance = str_to_real(word)
        call get_param(line,word2,at)
        !y_wall_distance = str_to_real(word)
        call set_rectangular_vat(x_distance=str_to_double(word1),&
             y_distance=str_to_double(word2))
     else if(str_equal('runs ',word)) then
        call get_param(line,word,at)
        runs = str_to_int(word)
     else if(str_equal('initial_time ',word)) then
        call get_param(line,word,at)
        time = str_to_double(word)
     else if(str_equal('number_of_pebbles ',word) .or. str_equal('number_of_spheres ',word)) then
        call get_param(line,word,at)
        number_of_pebbles = str_to_int(word)
        !$ chunk_size = max(number_of_pebbles/(8*omp_get_num_procs()),1)
     else if(str_equal('kinetic_friction ',word)) then 
        call get_param(line,word,at)
        kinetic_friction = str_to_double(word)
     else if(str_equal('static_friction ',word)) then
        print *,"ERROR: static_friction is deprecated.  use static_friction_new"
        print *,"or use pebbles6"
        stop
     else if(str_equal('alpha ',word)) then
        call get_param(line,word,at)
        alpha = str_to_double(word)
     else if(str_equal('dash_pot ',word)) then
        call get_param(line,word,at)
        normal_dash_pot_constant = str_to_double(word)
        transverse_dash_pot_constant = normal_dash_pot_constant
     else if(str_equal('dash_pot2 ',word)) then
        call get_param(line,word,at)
        normal_dash_pot_constant = str_to_double(word)
        call get_param(line,word,at)
        transverse_dash_pot_constant = str_to_double(word)
     else if(str_equal('initial_packing ',word)) then        
        call get_param(line,word,at)
        initial_packing = str_to_double(word)
     else if(str_equal('random_packing_method ',word)) then
        call get_param(line,word,at)
        initial_rand_packing_mult = str_to_int(word)
     else if(str_equal('dump_frequency ',word)) then
        call get_param(line,word,at)
        dump_frequency = str_to_int(word)
     else if(str_equal('display_frequency ',word)) then
        call get_param(line,word,at)
        display_frequency = str_to_int(word)
     else if(str_equal('position_display_frequency ',word)) then
        call get_param(line,word,at)
        position_frequency = str_to_int(word)
     else if(str_equal('energy_display_frequency ',word)) then
        call get_param(line,word,at)
        energy_frequency = str_to_int(word)
     else if(str_equal('torque_enable ',word)) then 
        !torque_enable = .true.
     else if(str_equal('torque_disable ',word)) then
        print *,"ERROR: torque cannot be disabled, use pebbles6"
        !torque_enable = .false.
     else if(str_equal('seed ',word)) then
        call get_param(line,word,at)
        seed = str_to_int(word)
     else if(str_equal('earthquake_enable ',word)) then
        call get_param(line,word,at)
        !max_number_of_earthquake_waves = str_to_int(word)
        call enable_earthquake_waves(str_to_int(word))
     else if(str_equal('earthquake_sine_wave ',word)) then
        call parse_earthquake_sine_wave_line(line,at)
     else if(str_equal('earthquake_sine_wave_offset ',word)) then
        call parse_eq_sine_wave_offset_line(line,at)
     else if(str_equal('dump_positions ',word)) then
        call get_param(line,word,at)
        dump_positions_filename = word
        do_dump_positions = .true.
     else if(str_equal('dump_positions_mult ',word)) then
        call get_param(line,word,at)
        dump_positions_filename = word
        call get_param(line,word,at)
        dump_positions_mult = str_to_double(word)
        do_dump_positions = .true.
     else if(str_equal('load_positions ',word)) then
        call get_param(line,word,at)
        load_positions_filename = word
        do_load_positions = .true.
     else if(str_equal('load_positions_divide ',word)) then
        call get_param(line,word,at)
        load_positions_filename = word
        call get_param(line,word,at)
        load_positions_divide = str_to_double(word)
        do_load_positions = .true.
     else if(str_equal('load_pebble_info ',word) .or. str_equal('load_sphere_info ',word)) then
        call get_param(line,word,at)
        load_pebble_info_filename = word
        do_load_pebble_info = .true.
     else if(str_equal('load_pebble_slips ',word) .or. str_equal('load_sphere_slips ',word)) then
        do_load_pebble_slips = .true.
     else if(str_equal('load_pebble_surface_slips ',word) .or. str_equal('load_sphere_surface_slips ',word)) then
        do_load_pebble_surface_slips = .true.
     else if(str_equal('load_pebble_save ',word) .or. str_equal('load_sphere_save ',word)) then
        call get_param(line,word,at)
        load_pebble_info_filename = word
        do_load_pebble_info = .true.
        do_load_pebble_slips = .true.
        do_load_pebble_surface_slips = .true.
     else if(str_equal('floor_location ',word)) then
        call get_param(line,word,at)
        call set_floor_location(str_to_double(word))
     else if(str_equal('pebble_pebble_hooke ',word) .or. str_equal('sphere_sphere_hooke ',word)) then
        call get_param(line,word,at)
        pebble_pebble_hooke = str_to_double(word)
     else if(str_equal('pebble_vessel_hooke ',word) .or. str_equal('sphere_vessel_hooke ',word)) then
        call get_param(line,word,at)
        pebble_vessel_hooke = str_to_double(word)
     else if(str_equal('pebble_radius ',word) .or. str_equal('sphere_radius ',word)) then
        call get_param(line,word,at)
        pebble_center_radius = str_to_double(word)
        call get_param(line,word,at)
        pebble_radius = str_to_double(word)
     else if(str_equal('pebble_density ',word) .or. str_equal('sphere_density ',word)) then
        call get_param(line,word,at)
        density_center = str_to_double(word)
        call get_param(line,word,at)
        density_outside = str_to_double(word)
     else if(str_equal('static_friction_new ',word)) then
        use_static_friction = .true.
        call get_param(line,word,at)
        pp_static_friction = str_to_double(word)
        surface_static_friction = pp_static_friction
        call get_param(line,word,at)
        static_friction_hooke = str_to_double(word)
        surface_static_friction_hooke = static_friction_hooke
        call get_param(line,word,at)
        static_velocity_sqr_max = str_to_double(word)
     else if(str_equal('static_friction_new2 ',word)) then
        use_static_friction = .true.
        call get_param(line,word,at)
        pp_static_friction = str_to_double(word)
        surface_static_friction = pp_static_friction
        call get_param(line,word,at)
        static_friction_hooke = str_to_double(word)
        call get_param(line,word,at)
        surface_static_friction_hooke = str_to_double(word)
        call get_param(line,word,at)
        static_velocity_sqr_max = str_to_double(word)
     else if(str_equal('static_friction_new3 ',word)) then
        use_static_friction = .true.
        call get_param(line,word,at)
        pp_static_friction = str_to_double(word)
        call get_param(line,word,at)
        static_friction_hooke = str_to_double(word)
        call get_param(line,word,at)
        surface_static_friction = str_to_double(word)
        call get_param(line,word,at)
        surface_static_friction_hooke = str_to_double(word)
        call get_param(line,word,at)
        static_velocity_sqr_max = str_to_double(word)
     else if(str_equal('zero_long_slips ',word)) then
        zero_long_slips = .true.
        call get_param(line,word,at)  
        long_slip_sqr = str_to_double(word)**2
     else if(str_equal('decrease_long_slips ',word)) then
        decrease_long_slips = .true.
        call get_param(line,word,at)  
        decrease_slip_mult = str_to_double(word)
        call get_param(line,word,at)  
        decrease_slip_slope = str_to_double(word)
     else if(str_equal('zero_long_slips_mult ',word)) then
        zero_long_slips = .true.
        call get_param(line,word,at)  
        long_slip_mult = str_to_double(word)
     else if(str_equal('shorten_long_slips ',word)) then
        shorten_long_slips = .true.
        call get_param(line,word,at)  
        long_slip_sqr = str_to_double(word)**2
     else if(str_equal('zero_static_slips ',word)) then
        zero_static_slips = .true.
     else if(str_equal('calculate_pebble_wear ',word) .or. str_equal('calculate_sphere_wear ',word)) then
        calculate_pebble_wear_vol = .true.
        call get_param(line,word,at)
        pebble_wear_vol_display_freq = str_to_int(word)
     else if(str_equal('normal_force_wear_exponent ',word)) then
        call get_param(line,word,at)
        call set_wear_force_exponent(str_to_double(word))
     else if(str_equal('wear_velocity_bins ',word)) then
        call get_param(line,word,at)
        allocate(temp_real_array(str_to_int(word)))
        do i=1,size(temp_real_array)
           call get_param(line,word,at)
           temp_real_array(i) = str_to_double(word)
        end do
        call set_wear_velocity_bins(temp_real_array)
        deallocate(temp_real_array)
     else if(str_equal('linear_static_friction_cutoff ',word)) then
        linear_static_velocity_clamp = .true.
     else if(str_equal('no_linear_static_friction_cutoff ',word)) then
        linear_static_velocity_clamp = .false.
     else if(str_equal('pebble_of_interest ',word) .or. str_equal('sphere_of_interest ',word)) then
        call get_param(line,word,at)
        pebble_of_interest = str_to_int(word)
     else if(str_equal('use_simple_integrator ',word)) then
        integration_method = euler_method
     else if(str_equal('integration_method ',word)) then
        call get_param(line,word,at)
        if(str_equal('euler ',word)) then
           integration_method = euler_method
        else if(str_equal('runge_kutta ',word)) then
           integration_method = runge_kutta_method
        else if(str_equal('adams_bashforth ',word)) then
           integration_method = adams_bashforth_method
        else if(str_equal('adams_moulton ',word)) then
           integration_method = adams_moulton_method
        else
           print *,"ERROR unknown integration method ",word
           stop
        end if
     else if(str_equal('cone ',word)) then
        call get_param(line,word1,at)        
        !cone_location = str_to_double(word1)
        call get_param(line,word2,at)
        call set_cone(cone_location_d=str_to_double(word1), &
             cone_slope_d=str_to_double(word2))
        !cone_slope = str_to_double(word2)
     else if(str_equal('recirculate ',word)) then
        recirculate = .true.
        call get_param(line,word,at)
        call set_recirculation_height(str_to_double(word))
        call get_param(line,word,at)
        door_closed_time = str_to_double(word)
        call get_param(line,word,at)
        first_door_open_time = str_to_double(word)
        call get_param(line,word4,at)
        !hole_size = str_to_double(word)
        call get_param(line,word5,at)
        !hole_depth = str_to_double(word)
        call set_hole(hole_size_d=str_to_double(word4),&
             hole_depth_d=str_to_double(word5))
     else if(str_equal('exit_chute ',word)) then
        call get_param(line,word1,at)
        !hole_size = str_to_double(word)
        call get_param(line,word2,at)
        !hole_depth = str_to_double(word)
        call set_hole(hole_size_d=str_to_double(word1),&
             hole_depth_d=str_to_double(word2))
     else if(str_equal('recirculate_params ',word)) then
        recirculate = .true.
        call get_param(line,word,at)
        call set_recirculation_height(str_to_double(word))
        call get_param(line,word,at)
        door_closed_time = str_to_double(word)
        call get_param(line,word,at)
        first_door_open_time = str_to_double(word)
     else if(str_equal('velocity_tally ',word)) then
        call get_param(line,word,at)        
        first_velocity_tally_time = str_to_double(word)
        call get_param(line,word,at)
        velocity_tally_frequency = str_to_int(word)
     else if(str_equal('probability_table ',word)) then
        run_probability_tally = .true.
        call get_param(line,word,at)
        probability_tally_start = str_to_double(word)
        call get_param(line,word,at)
        probability_tally_frequency = str_to_int(word)
        call get_param(line,word,at)
        probability_tally_delta = str_to_double(word)
     else if(str_equal('number_of_cones ',word)) then
        call get_param(line,word,at)
        !number_of_cones = str_to_int(word)
        call set_number_of_cones(str_to_int(word))
     else if(str_equal('tabular_earthquake ',word)) then
        call parse_tabular_earthquake(line,at)
     else if(str_equal('cone_packing ',word)) then
        !use_cone_packing = .true.
        call get_param(line,word1,at)
        !packing_cone_radius = str_to_double(word)     
        call get_param(line,word2,at)
        !packing_cone_bottom = str_to_double(word)     
        call setup_cone_packing(cone_radius = str_to_double(word1),&
             cone_bottom = str_to_double(word2),chute_length = 0.0_d_t)
     else if(str_equal('cone_packing_chute ',word)) then
        !use_cone_packing = .true.
        call get_param(line,word1,at)
        !packing_cone_radius = str_to_double(word)     
        call get_param(line,word2,at)
        !packing_cone_bottom = str_to_double(word)     
        call get_param(line,word3,at)
        !packing_chute_length = str_to_double(word)
        call setup_cone_packing(cone_radius = str_to_double(word1),&
             cone_bottom = str_to_double(word2), &
             chute_length = str_to_double(word3))
     else if(str_equal('offset_inlet_holes ',word)) then
        !offset_inlet_and_cone = .true.
        call set_offset_inlet_and_cone()
     else if(str_equal('sort_pebbles ',word) .or. str_equal('sort_spheres ',word)) then
        sort_pebbles = .true.
     else if(str_equal('calculate_force_tally ',word)) then
        calculate_force_tally = .true.
     else if(str_equal('ppf_filename ',word)) then
        print *,"#ppf_filename ignored"
     else if(str_equal('ppf_rewind_frequency ',word)) then
        print *,"#ppf_rewind_frequency ignored"
     else
        print *,"ERROR unknown keyword ",word
        stop
     end if
  end do input_loop

  ! Do calcuations of secondary parameters and allocations
  if(integration_method == euler_method &
       .or.  integration_method == runge_kutta_method)then 
     multistep_order = 2
  else if(integration_method == adams_bashforth_method &
       .or. integration_method == adams_moulton_method) then
     multistep_order = 4
  end if
  static_velocity_max = sqrt(static_velocity_sqr_max)
  long_slip = sqrt(long_slip_sqr)

  area = get_vessel_area()
  fill_edge = get_vessel_fill_edge()

  pebble_volume = (4.0/3.0)*pi*pebble_radius**3  
  height_needed = REAL(number_of_pebbles)*pebble_volume/area  
  mass = (4.0/3.0)*pi*(density_center*pebble_center_radius**3+&
       density_outside*(pebble_radius**3 - pebble_center_radius**3))
  gravitational_force = gravitational_acceleration*mass
  moment_of_inertia = 8.0/15.0 * pi * &
       (density_center * pebble_center_radius**5 + &
       density_outside * (pebble_radius ** 5 - pebble_center_radius ** 5))
  zone_increment = pebble_radius*2.0  
  call get_max_xy_extents(x_high,y_high)
  call initialize_geometry(z_low,z_high)
  z_high = z_high + 2*height_needed/initial_packing
  x_low = -x_high
  y_low = -y_high
  if(calculate_force_tally) then
     call initialize_force_tally(x_high,x_low,y_high,y_low,z_high,z_low)
  end if
  x_size = CEILING((x_high-x_low)/zone_increment)
  y_size = CEILING((y_high-y_low)/zone_increment)
  z_size = CEILING((z_high-z_low)/zone_increment)
  tally_z_inc = pebble_radius
  tally_radial_inc = pebble_radius
  rt_size = CEILING((y_high)/tally_radial_inc)
  zt_size = CEILING((z_high-z_low)/tally_z_inc)

  allocate(states(multistep_order),deltas(multistep_order))
  do i=1,multistep_order
     allocate(states(i)%pp(number_of_pebbles))
     allocate(states(i)%pv(number_of_pebbles))
     allocate(states(i)%pw(number_of_pebbles))
     ! If we don't allocate various fortran compilers act stupid.
     call allocate_pp_slip(number_of_pebbles,states(i)%pp_slip_data)
     allocate(states(i)%pss(number_of_pebbles))
     !allocate(states(i)%psn(number_of_pebbles))

     allocate(deltas(i)%pdp(number_of_pebbles))
     allocate(deltas(i)%pdv(number_of_pebbles))
     allocate(deltas(i)%pdw(number_of_pebbles))
     allocate(deltas(i)%pdss(number_of_pebbles))
     allocate(deltas(i)%keep_pss(number_of_pebbles))
     call allocate_pp_slip(number_of_pebbles,deltas(i)%pp_slip_delta)
  end do
  call initialize_pebble_cache(nearby_pebble_cache,x_size,x_low,y_size,y_low,&
       z_size,z_low,zone_increment)
  !allocate(zone(x_size,y_size,z_size,zone_count),stat=allocate_status)
  !if (allocate_status /= 0) stop "ERROR unable to allocate ram"
  !Allocate the radial tally zones
  allocate(zone_n(tally_sections,rt_size,zt_size))
  allocate(zone_vel_sum(tally_sections,rt_size,zt_size))
  allocate(zone_dot_sum(tally_sections,rt_size,zt_size))
  !Allocate the rectangular tally zones
  allocate(zone_rect_n(x_size,y_size,z_size),stat=allocate_status)
  if (allocate_status /= 0) stop "ERROR unable to allocate ram"
  allocate(zone_rect_vel_sum(x_size,y_size,z_size),stat=allocate_status)
  if (allocate_status /= 0) stop "ERROR unable to allocate ram"
  allocate(zone_rect_dot_sum(x_size,y_size,z_size),stat=allocate_status)
  if (allocate_status /= 0) stop "ERROR unable to allocate ram"

  call initialize_wear_tallies(z_low,zone_increment,z_size)
  if(run_probability_tally) then
     call initialize_count_data(probability_data,probability_tally_delta,& 
          x_low,x_high,y_low,y_high,z_low,z_high)
  end if

  time_till_door_open = first_door_open_time

  call initialize_earthquake

  !zero the tallies
  zone_n = 0
  zone_vel_sum = vector(0.0,0.0,0.0)
  zone_dot_sum = 0.0
  zone_rect_n = 0
  zone_rect_vel_sum = vector(0.0,0.0,0.0)
  zone_rect_dot_sum = 0.0


  if(use_static_friction) then 
     call clear_pp_slip(states(1)%pp_slip_data)
  end if

  call do_runs()

  ! deallocate memory
  do i=1,multistep_order
     deallocate(states(i)%pp,states(i)%pv,states(i)%pw)
     if(use_static_friction) then
        call deallocate_pp_slip(states(i)%pp_slip_data)
        deallocate(states(i)%pss) !,states(i)%psn)
     end if
     deallocate(deltas(i)%pdp,deltas(i)%pdv,deltas(i)%pdw)
  end do
  
  call free_earthquake
CONTAINS  

  subroutine print_parameters() 
    PRINT *,"#Pebbles"  
    PRINT *,"#Version $Id: pebbles9.f90 1378 2011-02-24 15:10:13Z cogljj $"    
    call print_version
    PRINT *,"# Prepared by Battelle Energy Alliance, LLC"
    PRINT *,"# Under Contract No. DE-AC07-05ID14517"
    PRINT *,"# With the U. S. Department of Energy"
    PRINT *,"#Pebble_Count: ",number_of_pebbles  
    PRINT *,"#alpha",alpha," grav",gravitational_force  
    PRINT *,"#runs",runs,runs*alpha,"s long",time,"s start"
    PRINT *,"#sizes",x_size,y_size,z_size  
    PRINT *,"#sp sf k friction",pp_static_friction,surface_static_friction,kinetic_friction  
    if(use_static_friction) then
       PRINT *,"#static velocity max",static_velocity_max
       PRINT *,"#static friction hooke",static_friction_hooke, surface_static_friction_hooke
       PRINT *,"#linear static velocity",linear_static_velocity_clamp,zero_static_slips
    else 
       PRINT *,"#Not using static friction"
    end if
    PRINT *,"#n_dashpot",normal_dash_pot_constant," t_dashpot ",&
         transverse_dash_pot_constant," air_friction ",air_friction  
    PRINT *,"#hooke", pebble_pebble_hooke, pebble_vessel_hooke  
    PRINT *,"#d_t",d_t
    PRINT *,"#display_frequency",display_frequency,display_frequency*alpha
    PRINT *,"#mass",mass," moment_of_inertia" ,moment_of_inertia
    call print_geometry_info()
    if(recirculate) then 
       PRINT *,"#door_closed_time",door_closed_time,"ttdo",time_till_door_open
    else 
       PRINT *,"#recirculation off"
    end if
    PRINT *,"#first_velocity_tally_time",first_velocity_tally_time,"vtf",&
         velocity_tally_frequency
    call print_earthquake_info
    PRINT *,"#torque_enable",torque_enable
    PRINT *,"#static_friction",use_static_friction,"spf",pp_static_friction,&
         "ssf",surface_static_friction,&
         "hooke",static_friction_hooke,"v**2",static_velocity_sqr_max,&
         "surface_hooke",surface_static_friction_hooke
    PRINT *,"#display_frequency",display_frequency,display_frequency*alpha
    PRINT *,"#run_length",runs,runs*alpha
    PRINT *,"#seed",seed
    PRINT *,"#use_simple_integrator",integration_method == euler_method
    print *,"#sizes",x_size,y_size,z_size
    print *,"#near_pebble_size",near_pebbles_size
    !$ print *,"#omp_procs",omp_get_num_procs(),omp_get_max_threads()
    call print_date("#start_time")
  end subroutine print_parameters

  subroutine do_runs()
    !t_c = pebble multistep time index current
    !t_m1 = pebble multistep index previous
    !t_m2 = pebble multistep index for current time - alpha * 2
    integer :: t_c,t_m1,t_m2,t_m3

    call print_parameters()

    call initialize_pebble_data()

    !     Run the data
    t_c = 1
    run_loop: DO i=0,runs     
       PRINT *,"#Run",i
       call print_time("Start loop")
       !t_m3 is index of time three steps back
       t_m3 = 1 + mod(i-2,multistep_order)
       t_m2 = 1 + mod(i-1,multistep_order)
       t_m1 = 1 + mod(i,multistep_order)
       !t_c is the current time.  states(t_c) is where the new data is written
       t_c = 1 + mod(i+1,multistep_order)
       call begin_new_force_tally_timestep()
       if(integration_method == euler_method) then 
          call get_simple_timestep(states(t_c),states(t_m1),deltas(t_c),&
               alpha,time)          
       else if(integration_method == runge_kutta_method .or. &

            (i <= 4 .and. (integration_method == adams_bashforth_method .or. &
            integration_method == adams_moulton_method))) then
          call get_rk_timestep(states(t_c),states(t_m1),deltas(t_c),alpha,time)
       else if(integration_method == adams_bashforth_method .and. i > 4) then
          call get_ab_timestep(states(t_c),states(t_m1),deltas,&
               t_c,t_m1,t_m2,t_m3,alpha,time)
       else if(integration_method == adams_moulton_method .and. i > 4) then
          call get_am_timestep(states(t_c),states(t_m1),deltas,&
               t_c,t_m1,t_m2,t_m3,alpha,time)
       else 
          print *,"ERROR, no integration method found for timestep",integration_method,i
          stop
       end if
       call handle_velocity_tally(states(t_c),time,i)
       if (recirculate) then 
          call handle_recirculation(states(t_c))
       end if

       if(run_probability_tally .and. time >= probability_tally_start &
            .and. MOD(i,probability_tally_frequency) .eq. 0) then
          call handle_probability_tally(states(t_c),states(t_m1))
       end if

       IF(MOD(i,display_frequency) .eq. 0 .or. i == runs) THEN
          PRINT *,"#Display Run",i
          call display_pebble_info(time,states(t_c))
       else if(mod(i,energy_frequency) .eq. 0) then
          call display_pebble_energy(time,states(t_c))
       END IF
       if(mod(i,position_frequency) .eq. 0) then
          call display_pebble_positions(time,states(t_c))
       end if
       if(mod(i,dump_frequency) .eq. 0 .or. i == runs) then
          call display_overlaps(states(t_c)%pp)
          call display_zone_velocities()
       end if
       if(any_pebbles_of_interest()) then
          call dump_pebbles_of_interest_info(states(t_c),deltas(t_c),time)
       end if
       if(calculate_pebble_wear_vol .and. mod(i,pebble_wear_vol_display_freq) == 0) then
          call print_pebble_wear_data(time,alpha)
       end if
       time = time + alpha
    END DO run_loop

    if(do_dump_positions) then
       call dump_positions(states(t_c)%pp,&
            dump_positions_filename,dump_positions_mult)
    end if

    if(run_probability_tally .and. time >= probability_tally_start) then
       call print_data(probability_data)
    end if

    if(calculate_force_tally) then
       call print_force_tally
    end if
    call print_date("#__end_time")
  end subroutine do_runs

END PROGRAM marbles9
