! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module nearby_pebble_cache_module
  use vector_type
  use number_utils
  use types
  use add_and_fetch_module
  !$ use omp_lib
  implicit none

  type nearby_pebble_cache_type
     private
     integer,pointer,dimension(:,:,:,:) :: zone
     !Number of pebbles in each zone.
     integer,pointer,dimension(:,:,:) :: num_in_zone
     logical,pointer, dimension(:) :: x_used,y_used,z_used
     integer :: x_size,y_size,z_size
     real(KIND=D_T):: x_low, y_low, z_low
     real(KIND=D_T) :: zone_increment
  end type nearby_pebble_cache_type

  private
  public :: nearby_pebbles, update_nearby_pebble_cache, initialize_pebble_cache
  public :: near_pebbles_size, nearby_pebble_cache_type
  public :: update_nearby_pebble_cache_s,add_to_nearby_pebble_cache
  public :: pebbles_in_zone, copy_pebbles_in_zone, get_next_zone
  public :: nearby_pebbles_all
  !Note: The use of pebbles_in_zone, copy_pebbles_in_zone and get_next_zone is 
  ! discouraged.
  integer, parameter :: zone_count = 20
  integer, parameter :: near_pebbles_size = 27*zone_count

contains

  subroutine initialize_pebble_cache(npc,x_size,x_low,y_size,y_low,&
       z_size,z_low,zone_increment)
    real(kind=d_t), intent(in) :: x_low,y_low,z_low,zone_increment
    integer, intent(in) :: x_size,y_size,z_size
    type(nearby_pebble_cache_type),intent(out) :: npc
    integer :: allocate_status
    
    npc%x_size = x_size
    npc%y_size = y_size
    npc%z_size = z_size

    allocate(npc%num_in_zone(x_size,y_size,z_size))
    allocate(npc%zone(zone_count,x_size,y_size,z_size),stat=allocate_status)
    if (allocate_status /= 0) stop "ERROR unable to allocate ram in nearby_pebble_cache"
    allocate(npc%x_used(x_size),npc%y_used(y_size),npc%z_used(z_size))

    call full_clear_zones(npc)
    
    npc%x_low = x_low
    npc%y_low = y_low
    npc%z_low = z_low
    npc%zone_increment = zone_increment


  end subroutine initialize_pebble_cache

  pure integer function find_base_zone(low, increment, size, value)  
    real(KIND=D_T), intent(in) :: low,increment,value  
    integer, intent(in) :: size
    integer :: zone  
    zone = find_zone(low, increment, size, value)  
    find_base_zone = iclamp(2,zone,size-1)  
  end function find_base_zone
  

  pure integer function find_zone(low, increment, size, value)  
    real(KIND=D_T), intent(in) :: low,increment,value  
    integer, intent(in) :: size 
    integer :: bin  
    bin = floor((value - low) / increment)  
    find_zone = iclamp(1,bin,size)  
  end function find_zone

  pure integer function pebbles_in_zone(npc,x,y,z) 
    integer,intent(in) :: x,y,z
    type(nearby_pebble_cache_type),intent(in) :: npc
    pebbles_in_zone = npc%num_in_zone(x,y,z)
  end function pebbles_in_zone

  subroutine copy_pebbles_in_zone(npc,x,y,z,out_array,out_array_start,magic)
    integer,intent(in) :: x,y,z,out_array_start,magic
    type(nearby_pebble_cache_type),intent(in) :: npc
    integer,pointer,dimension(:,:) :: out_array
    integer :: zone_pebbles,i
    zone_pebbles = pebbles_in_zone(npc,x,y,z)
    do i=1,zone_pebbles
       out_array(magic,out_array_start+i) = npc%zone(i,x,y,z)
    end do
  end subroutine copy_pebbles_in_zone

  pure subroutine get_next_zone(npc, x_c,y_c,z_c) 
    type(nearby_pebble_cache_type),intent(in) :: npc
    integer, intent(inout) :: x_c,y_c,z_c
50  continue
    x_c = x_c + 1
    if(x_c > npc%x_size) then
       x_c = 0
       y_c = y_c + 1
       goto 50
    end if
    if(y_c > npc%y_size) then
       x_c = 0
       y_c = 1
       z_c = z_c + 1
       goto 50
    end if
    if(z_c > npc%z_size) then 
       return
    end if
    if(.not.npc%x_used(x_c)) then 
       ! x_c = x_c + 1  Already done after 50
       goto 50
    end if
    if(.not.npc%y_used(y_c)) then
       x_c = 0
       y_c = y_c + 1
       goto 50
    end if
    if(.not.npc%z_used(z_c)) then
       x_c = 0
       y_c = 1
       z_c = z_c + 1
       goto 50
    end if
    
  end subroutine get_next_zone


  pure subroutine clear_zones(npc)
    !integer, intent(inout), dimension(:,:,:,:) :: zone
    !(x_size,y_size,z_size,zone_count)  
    type(nearby_pebble_cache_type),intent(inout) :: npc
    
    integer :: i,j,k
    do i=1,npc%x_size  
       if (npc%x_used(i)) then 
          do j=1,npc%y_size  
             if (npc%y_used(j)) then
                do k=1,npc%z_size  
                   if (npc%z_used(k)) then
                      npc%num_in_zone(i,j,k) = 0
                   end if
                end do
             end if
          end do
       end if
    end do

    npc%x_used = .false.
    npc%y_used = .false.
    npc%z_used = .false.

  end subroutine clear_zones

  pure subroutine full_clear_zones(npc)
    type(nearby_pebble_cache_type),intent(inout) :: npc

    npc%zone(:,:,:,:) = 0
    npc%num_in_zone(:,:,:) = 0
    npc%x_used(:) = .false.
    npc%y_used(:) = .false.
    npc%z_used(:) = .false.

  end subroutine full_clear_zones

  integer function fetch_zone_index(npc,x,y,z)
    integer,intent(in) :: x,y,z
    type(nearby_pebble_cache_type),intent(inout) :: npc
    call atomic_add_and_fetch(npc%num_in_zone(x,y,z),1,fetch_zone_index)
  end function fetch_zone_index

!!$  pure integer function free_zone_index(npc,x,y,z) 
!!$    integer,intent(in) :: x,y,z
!!$    type(nearby_pebble_cache_type),intent(in) :: npc
!!$    integer :: i
!!$    do i=1,zone_count
!!$       if(npc%zone(x,y,z,i) <= 0) then
!!$          free_zone_index = i
!!$          return
!!$       end if
!!$    end do
!!$    free_zone_index = zone_count
!!$  end function free_zone_index

  subroutine update_nearby_pebble_cache(npc, pp)
    type(nearby_pebble_cache_type),intent(inout) :: npc

    type(vector), intent(in),dimension(:) :: pp
    integer :: ii
    !$ integer :: chunk_size
    call clear_zones(npc)
    !$ chunk_size = max(size(pp)/(8*omp_get_num_procs()),1)
    !$OMP PARALLEL DO SHARED(npc) SCHEDULE(dynamic,chunk_size)
    do ii=1,size(pp)
       call add_to_nearby_pebble_cache(npc, pp, ii)
    end do

  end subroutine update_nearby_pebble_cache

  subroutine update_nearby_pebble_cache_s(npc, pp, &
       selected, selected_count)
    type(nearby_pebble_cache_type),intent(inout) :: npc
    !(x_size,y_size,z_size,zone_count)  
    type(vector), intent(in),dimension(:) :: pp
    integer, intent(in), dimension(:) :: selected
    integer, intent(in) :: selected_count
    integer :: ii, j  
    call clear_zones(npc)
    !TODO should use omp parallel
    do j=1,selected_count
       ii = selected(j)
       call add_to_nearby_pebble_cache(npc, pp, ii)
    end do

  end subroutine update_nearby_pebble_cache_s
    
  subroutine add_to_nearby_pebble_cache(npc, pp, index)
    type(nearby_pebble_cache_type),intent(inout) :: npc
    !(x_size,y_size,z_size,zone_count)  
    type(vector), intent(in),dimension(:) :: pp
    integer, intent(in) :: index
    integer :: i,x_z,y_z,z_z

    x_z = find_zone(npc%x_low, npc%zone_increment, npc%x_size, pp(index)%x)  
    y_z = find_zone(npc%y_low, npc%zone_increment, npc%y_size, pp(index)%y)  
    z_z = find_zone(npc%z_low, npc%zone_increment, npc%z_size, pp(index)%z)  
    npc%x_used(x_z) = .true. 
    npc%y_used(y_z) = .true.
    npc%z_used(z_z) = .true.
    i = fetch_zone_index(npc,x_z,y_z,z_z)
    if(i >= zone_count) then
       print *,"#FULL ZONE",x_z,y_z,z_z,index,i,pp(index)
       !return
    end if
    !if(i >= 10) then
    !   print *,"#NEAR FULL",x_z,y_z,z_z,index,i,pp(index)
    !end if
    npc%zone(i, x_z,y_z,z_z) = index
    
  end subroutine add_to_nearby_pebble_cache



  !     Find all the pebbles that are in the current zone, and
  !     in the immediatly surrounding areas.  fill_zone should
  !     have been called previously.
  !     j can be -1 if the position p is not a current pebble position.
  subroutine nearby_pebbles(npc,p, ret_array, ret_array_size, number_found,j)
    type(nearby_pebble_cache_type),intent(inout) :: npc
    type(vector), intent(in) :: p
    integer, intent(in) :: ret_array_size
    integer, intent(out) :: number_found  
    integer, intent(out) :: ret_array(ret_array_size)  
    integer :: x_b,y_b,z_b,x_z,y_z,z_z,j,l  
    integer :: next_index, num_in_zone

    !First, find all the pebbles in the relevant zones.
    next_index = 1  
    x_b = find_base_zone(npc%x_low, npc%zone_increment, npc%x_size, p%x)  
    y_b = find_base_zone(npc%y_low, npc%zone_increment, npc%y_size, p%y)  
    z_b = find_base_zone(npc%z_low, npc%zone_increment, npc%z_size, p%z)  
    do x_z=x_b-1,x_b+1  
       do y_z=y_b-1,y_b+1  
          do z_z=z_b-1,z_b+1  
             num_in_zone = npc%num_in_zone(x_z,y_z,z_z)
             if (num_in_zone > 0) then
                ret_array(next_index:next_index+num_in_zone-1) = npc%zone(1:num_in_zone,x_z,y_z,z_z)
                next_index = next_index + num_in_zone
             end if
          end do
       end do
    end do
    !Then, remove the current pebble
    zone_loop: do l=1,next_index-1
       if(ret_array(l) == j) then 
          ret_array(l) = ret_array(next_index-1)
          next_index = next_index - 1
          exit zone_loop
       end if
    end do zone_loop
    number_found = next_index - 1

  end subroutine nearby_pebbles

  !     Find all the pebbles that are in the current zone, and
  !     in the immediatly surrounding areas.  fill_zone should
  !     have been called previously.
  !     Includes the pebble p if p is in cache.
  subroutine nearby_pebbles_all(npc,p, ret_array, ret_array_size, number_found)
    type(nearby_pebble_cache_type),intent(inout) :: npc
    type(vector), intent(in) :: p
    integer, intent(in) :: ret_array_size
    integer, intent(out) :: number_found  
    integer, intent(out) :: ret_array(ret_array_size)  
    integer :: x_b,y_b,z_b,x_z,y_z,z_z
    integer :: next_index, num_in_zone
    !character(1024) :: io_temp_buffer

    !First, find all the pebbles in the relevant zones.
    next_index = 1  
    x_b = find_base_zone(npc%x_low, npc%zone_increment, npc%x_size, p%x)  
    y_b = find_base_zone(npc%y_low, npc%zone_increment, npc%y_size, p%y)  
    z_b = find_base_zone(npc%z_low, npc%zone_increment, npc%z_size, p%z)  
    do z_z=z_b-1,z_b+1  
       do y_z=y_b-1,y_b+1  
          do x_z=x_b-1,x_b+1  
             num_in_zone = npc%num_in_zone(x_z,y_z,z_z)
             ret_array(next_index:next_index+num_in_zone-1) = npc%zone(1:num_in_zone,x_z,y_z,z_z)
             next_index = next_index + num_in_zone
          end do
       end do
    end do
    number_found = next_index - 1

  end subroutine nearby_pebbles_all
  
end module nearby_pebble_cache_module
