! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module new_fortran_utils
  use, intrinsic :: iso_fortran_env
  implicit none

  logical :: output_suppress = .false.,flush_everything = .false.

  private
  public :: flushed_time, write_time, get_stdout_unit, get_settings_unit
  public :: output_suppress, flush_everything
contains
  
  subroutine flushed_time(unit,string)
    character(*),intent(in) :: string
    integer,intent(in) :: unit
    real :: time
    integer :: seconds, real_seconds, ms
    integer,dimension(8) :: date_time
    if(output_suppress) then
       return
    end if
    call cpu_time(time)
    call date_and_time(values=date_time)
    seconds = floor(time)
    ms = floor((time - seconds) * 1000)
    real_seconds = date_time(5)*3600+date_time(6)*60+date_time(7)
    write(unit,'(I7,A1,I3.3,A2,I7,A1,I3.3,A2,A)') seconds,".",ms,"s",real_seconds,".",date_time(8),"s ",trim(string)
    flush unit
  end subroutine flushed_time

  subroutine write_time(unit,string)
    character(*),intent(in) :: string
    integer,intent(in) :: unit
    real :: time
    integer :: seconds, real_seconds, ms
    integer,dimension(8) :: date_time
    if(output_suppress) then
       return
    end if
    call cpu_time(time)
    call date_and_time(values=date_time)
    seconds = floor(time)
    ms = floor((time - seconds) * 1000)
    real_seconds = date_time(5)*3600+date_time(6)*60+date_time(7)
    write(unit,'(I7,A1,I3.3,A2,I7,A1,I3.3,A2,A)') seconds,".",ms,"s",real_seconds,".",date_time(8),"s ",trim(string)
    if(flush_everything) then
       flush unit
    end if
  end subroutine write_time

  function get_stdout_unit()
    integer :: get_stdout_unit
    get_stdout_unit = output_unit
  end function get_stdout_unit  

  function get_settings_unit(use_stdin)
    integer :: get_settings_unit
    integer :: settings_unit, openstatus
    logical :: use_stdin
    character(128) :: filename
    if(use_stdin .and. (command_argument_count() /= 1)) then
       get_settings_unit = input_unit
       return 
    end if
    if(command_argument_count() == 1) then
       call get_command_argument(1,filename)
    else 
       filename = "settings"
    end if
    
    settings_unit = 209
    open(settings_unit,file=filename,action="READ",status="OLD",iostat=openstatus)
    if(openstatus > 0) then
       print *,"ERROR unable to open file ",filename
       stop
    end if
    get_settings_unit = settings_unit
  end function get_settings_unit
  
end module new_fortran_utils
