! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module number_utils
  ! Make sure you edit number_utils.f90.m4 rather than number_utils.f90
  ! 
  use types
  implicit none
  real(kind=d_t), parameter :: pi = 3.141592653589793238462643383_d_t
  
  type real_int_pair
     real(kind=d_t) :: r
     integer :: i
  end type real_int_pair

  interface increment
     module procedure increment_int
  end interface

  interface increment
     module procedure increment_double
  end interface

contains
  function get_realseconds()
    real(kind=d_t) :: get_realseconds
    integer,dimension(8) :: date_time
    call date_and_time(values=date_time)
    get_realseconds = date_time(5)*3600+date_time(6)*60+date_time(7)+date_time(8)/1000.0_d_t
  end function get_realseconds

  pure subroutine increment_int(value,adder)
    integer,intent(inout) :: value
    integer,intent(in) :: adder
    
    value = value + adder

  end subroutine increment_int

  pure subroutine increment_double(value,adder)
    real(kind=d_t),intent(inout) :: value
    real(kind=d_t),intent(in) :: adder
    
    value = value + adder

  end subroutine increment_double

  subroutine print_date(header)
    character(*), intent(in) :: header
    character(8) :: date
    character(10) :: time
    call date_and_time(date,time)
    print *,header," ",date," ",time
  end subroutine print_date

  subroutine print_time(header) 
    character(*), intent(in) :: header
    real :: time
    integer :: seconds, real_seconds, ms
    integer,dimension(8) :: date_time
    call cpu_time(time)
    call date_and_time(values=date_time)
    seconds = floor(time)
    ms = floor((time - seconds) * 1000)
    real_seconds = date_time(5)*3600+date_time(6)*60+date_time(7)
    print '(A40,I7,A1,I3.3,A2,I7,A1,I3.3,A2)',header,seconds,".",ms,"s",real_seconds,".",date_time(8),"s"
  end subroutine print_time

  pure integer function iclamp(low,value,high)  
    integer, intent(in) :: low,value,high  
    if(value .lt. low) then  
       iclamp = low  
    else if(high .lt. value) then  
       iclamp = high
    else
       iclamp = value
    end if
  end function iclamp

  pure real(kind=d_t) function clamp(low,value,high)  
    real(kind=d_t), intent(in) :: low,value,high  
    if(value .lt. low) then  
       clamp = low  
    else if(high .lt. value) then  
       clamp = high
    else
       clamp = value
    end if
  end function clamp

  subroutine swap_int(a,b)
    integer,intent(inout) :: a,b
    integer :: temp
    temp = a 
    a = b
    b = temp
  end subroutine swap_int
  
  function gcd(c, d)
    integer,intent(in) :: c,d
    integer :: a,b, t
    integer :: gcd
    a = abs(c)
    b = abs(d)
    do 
       if(b == 0) then
          gcd = a
          return
       end if
       t = b
       b = mod(a,b)
       a = t
    end do
    gcd = a
  end function gcd

  !index_in_sorted_X_array returns the index of the value passed in,
  ! or the lower bound - 1 of the array if it is not found
  !search_sorted_X_array returns the index of the value passed in, or 
  ! if it is not found, the index which is has a value higher that it.
  ! this may be the upper bound + 1 if the value is higher than any in the 
  ! array
  !Both expect that the array is sorted, and may fail in strange ways 
  ! if it is not.


  integer function index_in_sorted_int_array(array,value) 
    integer,intent(in) :: value
    integer,dimension(:),intent(in) :: array
    integer :: high_index, low_index, mid_index
    high_index = ubound(array,1)
    low_index = lbound(array,1)
    index_in_sorted_int_array = lbound(array,1) - 1
    if(array(low_index) == value) then
       index_in_sorted_int_array = low_index
       return
    end if
    if(array(high_index) == value) then
       index_in_sorted_int_array = high_index
       return
    end if
    do 
       mid_index = floor((low_index+high_index)/2.0)
       if(array(mid_index) == value) then
          index_in_sorted_int_array = mid_index
          return
       end if
       if(value < array(mid_index)) then
          if(high_index == mid_index) return
          high_index = mid_index
       else if(array(mid_index) < value) then
          if(low_index == mid_index) return
          low_index = mid_index
       end if
       if(low_index > high_index) return
    end do
  end function index_in_sorted_int_array
 
  subroutine search_sorted_int_array(array,value,index,found) 
    integer,intent(in) :: value
    integer,dimension(:),intent(in) :: array
    integer,intent(out) :: index
    logical,intent(out) :: found
    integer :: high_index, low_index, mid_index
    high_index = ubound(array,1)
    low_index = lbound(array,1)
    index = lbound(array,1) - 1
    found = .false.
    if(array(low_index) == value) then
       index = low_index
       found = .true.
       return
    end if
    if(array(high_index) == value) then
       index = high_index
       found = .true.
       return
    end if
    if(array(high_index) < value) then
       index = ubound(array,1) + 1
       found = .false.
       return 
    end if
    if(array(low_index) > value) then
       index = low_index
       found = .false.
       return
    end if
    if(.not. (array(low_index) < value .and. value < array(high_index))) then
       !This can probably only happen with NaN.
       index = low_index - 1
       found = .false.
       return
    end if
    do 
       mid_index = floor((low_index+high_index)/2.0)
       if(array(mid_index) == value) then
          index = mid_index
          found = .true.
          return
       end if
       if(value < array(mid_index)) then
          if(high_index == mid_index) then 
             index = mid_index
             return
          end if
          high_index = mid_index
       else if(array(mid_index) < value) then
          if(low_index == mid_index) then
             index = high_index
             return
          end if
          low_index = mid_index
       end if
       if(low_index > high_index) return
    end do
  end subroutine search_sorted_int_array

  integer function index_in_sorted_real_array(array,value) 
    real(kind=d_t),intent(in) :: value
    real(kind=d_t),dimension(:),intent(in) :: array
    integer :: high_index, low_index, mid_index
    high_index = ubound(array,1)
    low_index = lbound(array,1)
    index_in_sorted_real_array = lbound(array,1) - 1
    if(array(low_index) == value) then
       index_in_sorted_real_array = low_index
       return
    end if
    if(array(high_index) == value) then
       index_in_sorted_real_array = high_index
       return
    end if
    do 
       mid_index = floor((low_index+high_index)/2.0)
       if(array(mid_index) == value) then
          index_in_sorted_real_array = mid_index
          return
       end if
       if(value < array(mid_index)) then
          if(high_index == mid_index) return
          high_index = mid_index
       else if(array(mid_index) < value) then
          if(low_index == mid_index) return
          low_index = mid_index
       end if
       if(low_index > high_index) return
    end do
  end function index_in_sorted_real_array
 
  subroutine search_sorted_real_array(array,value,index,found) 
    real(kind=d_t),intent(in) :: value
    real(kind=d_t),dimension(:),intent(in) :: array
    integer,intent(out) :: index
    logical,intent(out) :: found
    integer :: high_index, low_index, mid_index
    high_index = ubound(array,1)
    low_index = lbound(array,1)
    index = lbound(array,1) - 1
    found = .false.
    if(array(low_index) == value) then
       index = low_index
       found = .true.
       return
    end if
    if(array(high_index) == value) then
       index = high_index
       found = .true.
       return
    end if
    if(array(high_index) < value) then
       index = ubound(array,1) + 1
       found = .false.
       return 
    end if
    if(array(low_index) > value) then
       index = low_index
       found = .false.
       return
    end if
    if(.not. (array(low_index) < value .and. value < array(high_index))) then
       !This can probably only happen with NaN.
       index = low_index - 1
       found = .false.
       return
    end if
    do 
       mid_index = floor((low_index+high_index)/2.0)
       if(array(mid_index) == value) then
          index = mid_index
          found = .true.
          return
       end if
       if(value < array(mid_index)) then
          if(high_index == mid_index) then 
             index = mid_index
             return
          end if
          high_index = mid_index
       else if(array(mid_index) < value) then
          if(low_index == mid_index) then
             index = high_index
             return
          end if
          low_index = mid_index
       end if
       if(low_index > high_index) return
    end do
  end subroutine search_sorted_real_array

  logical function less_than_or_equal_int(a,b) 
    integer,intent(in) :: a,b
    less_than_or_equal_int = (a <= b)
  end function less_than_or_equal_int

  logical function less_than_or_equal_real(a,b)
    real(kind=d_t),intent(in) :: a,b
    less_than_or_equal_real = (a <= b)
  end function less_than_or_equal_real

  logical function less_than_or_equal_ri_pair(a,b)
    type(real_int_pair),intent(in) :: a,b
    less_than_or_equal_ri_pair = (a%r <= b%r)
  end function less_than_or_equal_ri_pair
    

  subroutine sort_int(value_array) 
    integer,dimension(:),intent(inout) :: value_array
    integer,dimension(:),allocatable :: merge_array
    integer :: lower,upper
    lower = lbound(value_array,1)
    upper = ubound(value_array,1)
    allocate(merge_array(lower:upper))
    call sort_int_backend(value_array,merge_array)
    deallocate(merge_array)
  end subroutine sort_int

  recursive subroutine sort_int_backend(value_array,merge_array) 
    integer,dimension(:),intent(inout) :: value_array
    integer,dimension(:),intent(inout) :: merge_array
    integer :: lower, upper, middle
    integer :: low_index,high_index, merge_index
    integer :: temp
    lower = lbound(value_array,1)
    upper = ubound(value_array,1)
    !print *,"in sort_int",lower,upper,value_array
    if(upper - lower <= 0) then
       return
    else if(upper - lower == 1) then
       if(.not.(less_than_or_equal_int(value_array(lower),value_array(upper)))) then
          temp = value_array(upper)
          value_array(upper) = value_array(lower)
          value_array(lower) = temp
       end if
       return
    end if
    middle = (upper - lower) / 2  + lower
    call sort_int(value_array(lower:middle))
    call sort_int(value_array(middle+1:upper))

    low_index = lower
    high_index = middle+1
    merge_index = lower 
    do
       if(low_index > middle) then 
          merge_array(merge_index) = value_array(high_index)
          high_index = high_index + 1
       else if(high_index > upper) then
          merge_array(merge_index) = value_array(low_index)
          low_index = low_index + 1
       else if(less_than_or_equal_int(value_array(low_index),value_array(high_index))) then
          merge_array(merge_index) = value_array(low_index)
          low_index = low_index + 1
       else 
          merge_array(merge_index) = value_array(high_index)
          high_index = high_index + 1
       end if
       merge_index = merge_index + 1

       if(merge_index > upper) exit
    end do

    value_array(lower:upper) = merge_array(lower:upper)
        
  end subroutine sort_int_backend

  subroutine sort_real(value_array) 
    real(kind=d_t),dimension(:),intent(inout) :: value_array
    real(kind=d_t),dimension(:),allocatable :: merge_array
    integer :: lower,upper
    lower = lbound(value_array,1)
    upper = ubound(value_array,1)
    allocate(merge_array(lower:upper))
    call sort_real_backend(value_array,merge_array)
    deallocate(merge_array)
  end subroutine sort_real

  recursive subroutine sort_real_backend(value_array,merge_array) 
    real(kind=d_t),dimension(:),intent(inout) :: value_array
    real(kind=d_t),dimension(:),intent(inout) :: merge_array
    integer :: lower, upper, middle
    integer :: low_index,high_index, merge_index
    real(kind=d_t) :: temp
    lower = lbound(value_array,1)
    upper = ubound(value_array,1)
    !print *,"in sort_int",lower,upper,value_array
    if(upper - lower <= 0) then
       return
    else if(upper - lower == 1) then
       if(.not.(less_than_or_equal_real(value_array(lower),value_array(upper)))) then
          temp = value_array(upper)
          value_array(upper) = value_array(lower)
          value_array(lower) = temp
       end if
       return
    end if
    middle = (upper - lower) / 2  + lower
    call sort_real(value_array(lower:middle))
    call sort_real(value_array(middle+1:upper))

    low_index = lower
    high_index = middle+1
    merge_index = lower 
    do
       if(low_index > middle) then 
          merge_array(merge_index) = value_array(high_index)
          high_index = high_index + 1
       else if(high_index > upper) then
          merge_array(merge_index) = value_array(low_index)
          low_index = low_index + 1
       else if(less_than_or_equal_real(value_array(low_index),value_array(high_index))) then
          merge_array(merge_index) = value_array(low_index)
          low_index = low_index + 1
       else 
          merge_array(merge_index) = value_array(high_index)
          high_index = high_index + 1
       end if
       merge_index = merge_index + 1

       if(merge_index > upper) exit
    end do

    value_array(lower:upper) = merge_array(lower:upper)
        
  end subroutine sort_real_backend

  subroutine sort_ri_pair(value_array) 
    type(real_int_pair),dimension(:),intent(inout) :: value_array
    type(real_int_pair),dimension(:),allocatable :: merge_array
    integer :: lower,upper
    lower = lbound(value_array,1)
    upper = ubound(value_array,1)
    allocate(merge_array(lower:upper))
    call sort_ri_pair_backend(value_array,merge_array)
    deallocate(merge_array)
  end subroutine sort_ri_pair

  recursive subroutine sort_ri_pair_backend(value_array,merge_array) 
    type(real_int_pair),dimension(:),intent(inout) :: value_array
    type(real_int_pair),dimension(:),intent(inout) :: merge_array
    integer :: lower, upper, middle
    integer :: low_index,high_index, merge_index
    type(real_int_pair) :: temp
    lower = lbound(value_array,1)
    upper = ubound(value_array,1)
    !print *,"in sort_int",lower,upper,value_array
    if(upper - lower <= 0) then
       return
    else if(upper - lower == 1) then
       if(.not.(less_than_or_equal_ri_pair(value_array(lower),value_array(upper)))) then
          temp = value_array(upper)
          value_array(upper) = value_array(lower)
          value_array(lower) = temp
       end if
       return
    end if
    middle = (upper - lower) / 2  + lower
    call sort_ri_pair(value_array(lower:middle))
    call sort_ri_pair(value_array(middle+1:upper))

    low_index = lower
    high_index = middle+1
    merge_index = lower 
    do
       if(low_index > middle) then 
          merge_array(merge_index) = value_array(high_index)
          high_index = high_index + 1
       else if(high_index > upper) then
          merge_array(merge_index) = value_array(low_index)
          low_index = low_index + 1
       else if(less_than_or_equal_ri_pair(value_array(low_index),value_array(high_index))) then
          merge_array(merge_index) = value_array(low_index)
          low_index = low_index + 1
       else 
          merge_array(merge_index) = value_array(high_index)
          high_index = high_index + 1
       end if
       merge_index = merge_index + 1

       if(merge_index > upper) exit
    end do

    value_array(lower:upper) = merge_array(lower:upper)
        
  end subroutine sort_ri_pair_backend

end module number_utils
