! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module str_utils
  implicit none
  integer, private, parameter :: d_t = selected_real_kind(12,5)

  type string_node
     private 
     type(string_node),pointer :: next
     character,pointer,dimension(:) :: v
  end type string_node

  type string_queue
     private 
     type(string_node),pointer :: first => null(),last => null()
  end type string_queue

contains
  
  subroutine getword(line, word, start, length, next_loc, skip_whitespace)
    character(*), intent(in) :: line
    character(*), intent(out) :: word
    integer, intent(in), optional :: start   
    integer, intent(out), optional :: length
    integer, intent(inout), optional :: next_loc
    logical, intent(in), optional :: skip_whitespace
    logical :: skip_white
    integer :: i,j, offset 
    if (present(next_loc)) then 
       offset = next_loc - 1
    else if (.not. present(start)) then 
       offset = 0
    else 
       offset = start - 1
    end if
    if (present(skip_whitespace)) then
       skip_white = skip_whitespace
    else 
       skip_white = .false.
    end if
    do j=1,len(word)
       word(j:j) = ' '
    end do
    i = 1
    if( skip_white )then
       do 
          if(ichar(line(i+offset:i+offset)) .ne. ichar(' ')) exit
          offset = offset + 1
       end do
    end if
    do 
       if(ichar(line(i+offset:i+offset)) .eq. ichar(' ')) exit
       word(i:i) = line(i+offset:i+offset)
       i = i + 1
    end do
    if (present(next_loc)) then
       next_loc = offset + i + 1
    end if
    if (present(length)) then 
       length = i
    end if
  end subroutine getword

  subroutine get_param(line, word, next_loc) 
    character(*), intent(in) :: line
    character(*), intent(out) :: word
    integer, intent(inout) :: next_loc
    integer :: length
    call getword(line,word,length=length,next_loc=next_loc,&
         skip_whitespace=.true.)
    !print *,"length",length,line
    if(length < 2 .or. 50 < length) then
       print *,"ERROR: Missing parameter in line:",line
    end if
  end subroutine get_param

  pure function digit_value(c) 
    character(1), intent(in) :: c
    integer :: digit_value
    digit_value = ichar(c) - ichar('0')
  end function digit_value

  pure function str_to_int(foo) 
    character(*), intent(in) :: foo
    integer :: i, value, cv
    integer :: str_to_int
    integer :: negative
    value = 0
    i = 1
    if(digit_value(foo(i:i)) .eq. digit_value('-')) then
       negative = -1
       i = i + 1
    else 
       negative = 1
    end if
    do 
       cv = digit_value(foo(i:i))
       if( cv < 0 .or. 9 < cv) EXIT
       value = value * 10 + cv
       i = i + 1
    end do
    str_to_int = value * negative
  end function str_to_int

  pure function str_to_real(foo)
    character(*), intent(in) :: foo
    integer :: i, cv
    real ::  value, mult
    real :: str_to_real
    integer :: negative
    value = 0.0
    i = 1
    if(digit_value(foo(i:i)) .eq. digit_value('-')) then
       negative = -1
       i = i + 1
    else 
       negative = 1
    end if
    do 
       cv = digit_value(foo(i:i))
       if( cv < 0 .or. 9 < cv) EXIT
       value = value * 10 + cv
       i = i + 1
    end do
    mult = 0.1
    if(cv == digit_value('.')) then
       i = i + 1
       do 
          cv = digit_value(foo(i:i)) 
          if( cv < 0 .or. 9 < cv) EXIT
          value = value + cv * mult
          i = i + 1
          mult = mult * 0.1
       end do
    end if
    str_to_real = value * negative
  end function str_to_real

  pure function str_to_double(foo)
    character(*), intent(in) :: foo
    real(kind=d_t) :: str_to_double
    read (foo,*) str_to_double
  end function str_to_double

  pure function str_equal(a, b) 
    character(*), intent(in) :: a,b
    logical :: str_equal
    integer :: i, search_len
    search_len = min(len(a),len(b))
    str_equal = .false.
    do i=1,search_len
       if( a(i:i) .ne. b(i:i)) then
          exit
       end if
       if( a(i:i) .eq. ' ') then 
          str_equal = .true.
          exit
       end if
       if( i .eq. search_len .and. len(a) .eq. len(b))then 
          str_equal = .true.
       end if
    end do
  end function str_equal

  pure function str_array_to_char(v)
    character,pointer,dimension(:) :: v
    character(len=size(v)) :: str_array_to_char
    integer :: i
    do i=1,size(v)
       str_array_to_char(i:i) = v(i+lbound(v,1)-1)
    end do
  end function str_array_to_char

  pure function str_char_to_array(s)
    character,pointer,dimension(:) :: str_char_to_array
    character(*),intent(in) :: s
    integer :: i
    allocate(str_char_to_array(len(s)))
    do i=1,len(s)
       str_char_to_array(i) = s(i:i)
    end do
  end function str_char_to_array

  pure function ca_starts_with(s,v)
    character(*),intent(in) :: s
    character,pointer,dimension(:) :: v
    logical :: ca_starts_with
    integer :: i
    if(len(s) > size(v)) then
       ca_starts_with = .false.
       return
    end if
    do i=1,len(s)
       if(s(i:i) /= v(i)) then
          ca_starts_with = .false.
          return
       end if
    end do
    ca_starts_with = .true.
  end function ca_starts_with

  subroutine initialize_sq(sq)
    type(string_queue),intent(out) :: sq
    nullify(sq%first)
    nullify(sq%last)
  end subroutine initialize_sq

  subroutine sq_append(sq,s)
    character,pointer,dimension(:) :: s
    type(string_queue),intent(inout) :: sq
    type(string_node),pointer :: new_node
    allocate(new_node)
    new_node%v => s
    nullify(new_node%next)
    if(.not. associated(sq%first))  then
       sq%first => new_node
       sq%last => new_node
    else
       sq%last%next => new_node
       sq%last => new_node
    end if
  end subroutine sq_append

  subroutine sq_first(sq,first)
    type(string_queue),intent(in) :: sq
    type(string_node),pointer :: first  
    first => sq%first
  end subroutine sq_first

  subroutine sq_last(sq,last)
    type(string_queue),intent(in) :: sq
    type(string_node),pointer :: last  
    last => sq%last
  end subroutine sq_last

  subroutine sq_next(current,next)
    type(string_node),pointer :: current,next
    next => current%next
  end subroutine sq_next

  subroutine sn_value(current,v)
    type(string_node),pointer :: current
    character,pointer,dimension(:) :: v
    v => current%v
  end subroutine sn_value

  subroutine sq_inplace_next(current)
    type(string_node),pointer :: current
    current => current%next
  end subroutine sq_inplace_next

  subroutine sq_clear(sq)
    type(string_queue),intent(inout) :: sq
    type(string_node),pointer :: temp, next
    next => sq%first
    do 
       if(.not. associated(next)) exit
       temp => next
       next => next%next
       deallocate(temp)
    end do
    nullify(sq%first)
    nullify(sq%last)
  end subroutine sq_clear


end module str_utils
